# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mean field proximal policy optimaztion algorithm.

Reference:

    Algumaei, Talal, et al. "Regularization of the policy updates for
    stabilizing
    Mean Field Games." Pacific-Asia Conference on Knowledge Discovery and Data
    Mining. Cham: Springer Nature Switzerland, 2023. Available at:
    https://link.springer.com/chapter/10.1007/978-3-031-33377-4_28
"""

# pylint: disable=consider-using-from-import
# pylint: disable=g-importing-member

import numpy as np
import torch
from torch.distributions.categorical import Categorical
import torch.nn as nn
import torch.nn.functional as F

from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.algorithms.nash_conv import NashConv


class NashC(NashConv):
  """Mainly used to calculate the exploitability."""

  def __init__(self, game, distrib, pi_value, root_state=None):
    self._game = game
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self._distrib = distrib
    self._pi_value = pi_value
    self._br_value = best_response_value.BestResponse(
        self._game,
        self._distrib,
        value.TabularValueFunction(self._game),
        root_state=root_state,
    )


class Agent(nn.Module):
  """Mainly used to calculate the exploitability."""

  def __init__(self, info_state_size, num_actions):
    super(Agent, self).__init__()
    self.num_actions = num_actions
    self.info_state_size = info_state_size
    self.critic = nn.Sequential(
        self.layer_init(nn.Linear(info_state_size, 128)),
        nn.Tanh(),
        self.layer_init(nn.Linear(128, 128)),
        nn.Tanh(),
        self.layer_init(nn.Linear(128, 1)),
    )
    self.actor = nn.Sequential(
        self.layer_init(nn.Linear(info_state_size, 128)),
        nn.Tanh(),
        self.layer_init(nn.Linear(128, 128)),
        nn.Tanh(),
        self.layer_init(nn.Linear(128, num_actions)),
    )

  def layer_init(self, layer, bias_const=0.0):
    """Used to initalize layers."""
    nn.init.xavier_normal_(layer.weight)
    nn.init.constant_(layer.bias, bias_const)
    return layer

  def get_value(self, x):
    """Get the value of the state."""
    return self.critic(x)

  def get_action_and_value(self, x, action=None):
    """Get the action and value of the state."""
    logits = self.actor(x)
    probs = Categorical(logits=logits)
    if action is None:
      action = probs.sample()
    return action, probs.log_prob(action), probs.entropy(), self.critic(x)


class Policy(policy_std.Policy):
  """Required obeject to work with OpenSpiel.

  Used in updating the distribution using the policy nd in calculating the
  nash-convergance.
  """

  def __init__(self, game, agent, player_ids, device):
    super().__init__(game, player_ids)
    self.agent = agent
    self.device = device

  def action_probabilities(self, state, player_id=None):
    """Calculate the action probabilities of the state."""
    obs = torch.Tensor(state.observation_tensor()).to(self.device)
    legal_actions = state.legal_actions()
    logits = self.agent.actor(obs).detach().cpu()
    legat_logits = np.array([logits[action] for action in legal_actions])
    probs = np.exp(legat_logits - legat_logits.max())
    probs /= probs.sum(axis=0)

    # returns a dict with actions as keys and their probabilities as values
    return {
        action: probs[legal_actions.index(action)] for action in legal_actions
    }


def rollout(env, iter_agent, eps_agent, num_epsiodes, steps, device):
  """Generates num_epsiodes rollouts."""
  info_state = torch.zeros((steps, iter_agent.info_state_size), device=device)
  actions = torch.zeros((steps,), device=device)
  logprobs = torch.zeros((steps,), device=device)
  rewards = torch.zeros((steps,), device=device)
  dones = torch.zeros((steps,), device=device)
  values = torch.zeros((steps,), device=device)
  entropies = torch.zeros((steps,), device=device)
  t_actions = torch.zeros((steps,), device=device)
  t_logprobs = torch.zeros((steps,), device=device)

  step = 0
  for _ in range(num_epsiodes):
    time_step = env.reset()
    while not time_step.last():
      obs = time_step.observations["info_state"][0]
      obs = torch.Tensor(obs).to(device)
      info_state[step] = obs
      with torch.no_grad():
        t_action, t_logprob, _, _ = iter_agent.get_action_and_value(obs)
        action, logprob, entropy, ivalue = eps_agent.get_action_and_value(obs)

      time_step = env.step([action.item()])

      # iteration policy data
      t_logprobs[step] = t_logprob
      t_actions[step] = t_action

      # episode policy data
      logprobs[step] = logprob
      dones[step] = time_step.last()
      entropies[step] = entropy
      values[step] = ivalue
      actions[step] = action
      rewards[step] = torch.Tensor(time_step.rewards).to(device)
      step += 1

  history = {
      "info_state": info_state,
      "actions": actions,
      "logprobs": logprobs,
      "rewards": rewards,
      "dones": dones,
      "values": values,
      "entropies": entropies,
      "t_actions": t_actions,
      "t_logprobs": t_logprobs,
  }
  return history


def calculate_advantage(gamma, norm, rewards, values, dones, device):
  """Function used to calculate the Generalized Advantage estimate."""
  with torch.no_grad():
    next_done = dones[-1]
    next_value = values[-1]
    steps = len(values)
    returns = torch.zeros_like(rewards).to(device)
    for t in reversed(range(steps)):
      if t == steps - 1:
        nextnonterminal = 1.0 - next_done
        next_return = next_value
      else:
        nextnonterminal = 1.0 - dones[t + 1]
        next_return = returns[t + 1]
      returns[t] = rewards[t] + gamma * nextnonterminal * next_return

    advantages = returns - values

  if norm:
    advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

  return advantages, returns


def learn(
    history,
    optimizer_actor,
    optimize_critic,
    agent,
    num_minibatches=5,
    update_epochs=5,
    itr_eps=0.05,
    eps_eps=0.2,
    alpha=0.5,
    ent_coef=0.01,
    max_grad_norm=5,
):
  """Update the agent network (actor and critic)."""
  v_loss = None
  batch_size = history["actions"].shape[0]
  b_inds = np.arange(batch_size)
  mini_batch_size = batch_size // num_minibatches
  # get batch indices
  np.random.shuffle(b_inds)
  for _ in range(update_epochs):
    for start in range(0, batch_size, mini_batch_size):
      end = start + mini_batch_size
      mb_inds = b_inds[start:end]
      # for each update epoch shuffle the batch indices
      # generate the new logprobs, entropy and value then calculate the ratio
      b_obs = history["info_state"][mb_inds]
      b_advantages = history["advantages"][mb_inds]

      # Get the data under the episode policy (representative agent current
      # policy)
      _, newlogprob, entropy, new_value = agent.get_action_and_value(
          b_obs, history["actions"][mb_inds]
      )
      logratio = newlogprob - history["logprobs"][mb_inds]
      ratio = torch.exp(logratio)

      # Get the data under the iteration policy (the population policy)
      _, t_newlogprob, _, _ = agent.get_action_and_value(
          b_obs, history["t_actions"][mb_inds]
      )
      t_logratio = t_newlogprob - history["t_logprobs"][mb_inds]
      t_ratio = torch.exp(t_logratio)

      # iteration update PPO
      t_pg_loss1 = b_advantages * t_ratio
      t_pg_loss2 = b_advantages * torch.clamp(t_ratio, 1 - itr_eps, 1 + itr_eps)

      # episodic update PPO
      pg_loss1 = b_advantages * ratio
      pg_loss2 = b_advantages * torch.clamp(ratio, 1 - eps_eps, 1 + eps_eps)

      # Calculate the loss using our loss function
      pg_loss = (
          -alpha * torch.min(pg_loss1, pg_loss2).mean()
          - (1 - alpha) * torch.min(t_pg_loss1, t_pg_loss2).mean()
      )
      v_loss = F.smooth_l1_loss(
          new_value.reshape(-1), history["returns"][mb_inds]
      ).mean()
      entropy_loss = entropy.mean()
      loss = pg_loss - ent_coef * entropy_loss

      # Actor update
      optimizer_actor.zero_grad()
      loss.backward()
      nn.utils.clip_grad_norm_(agent.actor.parameters(), max_grad_norm)
      optimizer_actor.step()

      # Critic update
      optimize_critic.zero_grad()
      v_loss.backward()
      nn.utils.clip_grad_norm_(agent.critic.parameters(), max_grad_norm)
      optimize_critic.step()

  assert v_loss is not None
  return v_loss


def calculate_explotability(game, distrib, policy):
  """This function is used to log the results to tensor board."""
  initial_states = game.new_initial_states()
  pi_value = policy_value.PolicyValue(
      game, distrib, policy, value.TabularValueFunction(game)
  )
  m = {
      f"ppo_br/{state}": pi_value.eval_state(state) for state in initial_states
  }
  nashc = NashC(game, distrib, pi_value).nash_conv()
  m["nash_conv_ppo"] = nashc

  return m



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Randomly searches for a Restricted Nash Equilibrium.

"""

from typing import Optional

import cma
import numpy as np
from open_spiel.python.mfg.algorithms import utils
from open_spiel.python.mfg.algorithms.regret import regret_minimizer


def softmax(x):
  e = np.exp(x - np.max(x))
  return e / np.sum(e, axis=-1, keepdims=True)


class NashCMAES(regret_minimizer.RegretMinimizer):
  """Base class for Regret Minimizers.

  Implements base functions for regret minimizers to implement.

  Attributes:
    _game: Pyspiel game.
    _regret_steps_per_step: Number of regret steps per `step` call (Maximum
      number in case `stop_early` is true)
    _rho_tol: If `_compress_nus` is true, minimum probability threshold (
      Probabilities below `rho_tol` will be filtered out).
    _compress_nus: Whether to compress nus (Remove nus with low selection
      probability) or not.
    _compress_lbd: Penalty term in L1 minimization when compressing nus.
    _stop_early: Whether to stop regret computation when average regret is lower
      than `_stop_regret_threshold` or to keep going until
      `_regret_steps_per_step` steps have been accomplished.
    _stop_regret_threshold: If `stop_early` is true, average regret threshold
      under which the algorithm will stop.
    _policies: List of Policies
    _value_estimator: Value estimation function.
    _value_estimation_n: Number of runs to average _value_estimator's result on.
  """

  def __init__(self,
               game,
               policies,
               eta: Optional[float] = None,
               regret_steps_per_step: int = 1,
               rho_tol: float = 1e-4,
               compress_nus: bool = True,
               compress_lbd: float = 0.0,
               compress_every: int = 1,
               stop_early: bool = True,
               stop_regret_threshold: float = 1e-3,
               value_estimator=utils.sample_value,
               value_estimation_n: int = 1):
    super().__init__(
        game,
        policies,
        regret_steps_per_step=regret_steps_per_step,
        rho_tol=rho_tol,
        compress_nus=compress_nus,
        compress_every=compress_every,
        compress_lbd=compress_lbd,
        stop_early=stop_early,
        stop_regret_threshold=stop_regret_threshold,
        value_estimator=value_estimator,
        value_estimation_n=value_estimation_n)
    self._nu = np.ones(len(policies)) / len(policies)
    self._exploitability = None

  def compute_exploitability(self, nu):
    mu = utils.MixedDistribution(self._policy_mus, nu)
    per_policy_reward = 0.0
    for _ in range(self._value_estimation_n):
      per_policy_reward += np.array(
          [self._value_estimator(pi, mu, self._game) for pi in self._policies])
    per_policy_reward /= self._value_estimation_n
    on_policy_reward = np.sum(per_policy_reward * nu)
    return np.max(per_policy_reward - on_policy_reward)

  def step_for(self, T):  # pylint: disable=invalid-name
    self.step(T)

  def get_exploitabilities(self, nus):
    return np.array([self.compute_exploitability(nu) for nu in nus])

  def step(self, T):  # pylint: disable=invalid-name
    best_nu = np.ones(len(self._policies)) / len(self._policies)
    nu = best_nu
    n = 0
    best_exploitability = self.compute_exploitability(nu)
    exploitability = best_exploitability

    optimizer = cma.CMAEvolutionStrategy(x0=nu, sigma0=1.0)

    while best_exploitability > self._rho_tol and n < max(
        T, self._regret_steps_per_step):
      n += 1

      logit_nus = optimizer.ask()
      nus = softmax(logit_nus)
      exploitabilities = self.get_exploitabilities(nus)
      optimizer.tell(logit_nus, exploitabilities)

      best_new_exploitability = np.min(exploitabilities[0])
      if best_new_exploitability < best_exploitability:
        best_exploitability = best_new_exploitability
        best_nu = nus[np.argmin(exploitabilities)]
        print(best_exploitability)

    self._nus = [best_nu]
    self._nu_weights = [1.0]
    self._exploitability = exploitability

  def compute_average_regret(self):
    return self._exploitability

  def reset(self, policies):
    """Restart the bandit with new policies."""
    self._policies = policies
    self._policy_mus = []
    self._nu_weights = []
    self._exploitability = None
    self.update_policy_mus()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DQN agent implemented in JAX."""

import collections

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax
import rlax

from open_spiel.python import rl_agent
from open_spiel.python.utils.replay_buffer import ReplayBuffer

Transition = collections.namedtuple(
    "Transition",
    "info_state action reward next_info_state is_final_step legal_actions_mask")

# Penalty for illegal actions in action selection. In epsilon-greedy, this will
# prevent them from being selected.
ILLEGAL_ACTION_LOGITS_PENALTY = -1e9


class DQN(rl_agent.AbstractAgent):
  """DQN Agent implementation in JAX."""

  def __init__(self,
               player_id,
               state_representation_size,
               num_actions,
               hidden_layers_sizes=128,
               replay_buffer_capacity=10000,
               batch_size=128,
               replay_buffer_class=ReplayBuffer,
               learning_rate=0.01,
               update_target_network_every=1000,
               learn_every=10,
               discount_factor=1.0,
               min_buffer_size_to_learn=1000,
               epsilon_start=1.0,
               epsilon_end=0.1,
               epsilon_decay_duration=int(1e6),
               optimizer_str="sgd",
               loss_str="mse",
               huber_loss_parameter=1.0,
               seed=42,
               gradient_clipping=None):
    """Initialize the DQN agent."""

    # This call to locals() is used to store every argument used to initialize
    # the class instance, so it can be copied with no hyperparameter change.
    self._kwargs = locals()

    self.player_id = player_id
    self._num_actions = num_actions
    if isinstance(hidden_layers_sizes, int):
      hidden_layers_sizes = [hidden_layers_sizes]
    self._layer_sizes = hidden_layers_sizes
    self._batch_size = batch_size
    self._update_target_network_every = update_target_network_every
    self._learn_every = learn_every
    self._min_buffer_size_to_learn = min_buffer_size_to_learn
    self._discount_factor = discount_factor
    self.huber_loss_parameter = huber_loss_parameter

    self._epsilon_start = epsilon_start
    self._epsilon_end = epsilon_end
    self._epsilon_decay_duration = epsilon_decay_duration

    # TODO(author6) Allow for optional replay buffer config.
    if not isinstance(replay_buffer_capacity, int):
      raise ValueError("Replay buffer capacity not an integer.")
    self._replay_buffer = replay_buffer_class(replay_buffer_capacity)
    self._prev_timestep = None
    self._prev_action = None

    # Step counter to keep track of learning, eps decay and target network.
    self._step_counter = 0

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    # Create the Q-network instances

    def network(x):
      mlp = hk.nets.MLP(self._layer_sizes + [num_actions])
      return mlp(x)

    self.hk_network = hk.without_apply_rng(hk.transform(network))
    self.hk_network_apply = jax.jit(self.hk_network.apply)

    rng = jax.random.PRNGKey(seed)
    self._create_networks(rng, state_representation_size)

    if loss_str == "mse":
      self.loss_func = lambda x: jnp.mean(x**2)
    elif loss_str == "huber":
      # pylint: disable=g-long-lambda
      self.loss_func = lambda x: jnp.mean(
          rlax.huber_loss(x, self.huber_loss_parameter))
    else:
      raise ValueError("Not implemented, choose from 'mse', 'huber'.")

    if optimizer_str == "adam":
      optimizer = optax.adam(learning_rate)
    elif optimizer_str == "sgd":
      optimizer = optax.sgd(learning_rate)
    else:
      raise ValueError("Not implemented, choose from 'adam' and 'sgd'.")

    # Clipping the gradients prevent divergence and allow more stable training.
    if gradient_clipping:
      optimizer = optax.chain(optimizer,
                              optax.clip_by_global_norm(gradient_clipping))

    opt_init, opt_update = optimizer.init, optimizer.update

    self._opt_update_fn = self._get_update_func(opt_update)
    self._opt_state = opt_init(self.params_q_network)
    self._loss_and_grad = jax.value_and_grad(self._loss, has_aux=False)
    self._jit_update = jax.jit(self.get_update())

  def _create_networks(self, rng, state_representation_size):
    """Called to create the networks."""
    x = jnp.ones([1, state_representation_size])
    self.params_q_network = self.hk_network.init(rng, x)
    self.params_target_q_network = self.hk_network.init(rng, x)

  def _get_update_func(self, opt_update):

    def update(params, opt_state, gradient):
      """Learning rule (stochastic gradient descent)."""
      updates, opt_state = opt_update(gradient, opt_state)
      new_params = optax.apply_updates(params, updates)
      return new_params, opt_state

    return update

  def _get_action_probs(self, info_state, legal_actions, is_evaluation=False):
    """Returns a selected action and the probabilities of legal actions."""
    epsilon = self._get_epsilon(is_evaluation)
    return self._epsilon_greedy(info_state, legal_actions, epsilon)

  def step(self, time_step, is_evaluation=False, add_transition_record=True):
    """Returns the action to be taken and updates the Q-network if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.
      add_transition_record: Whether to add to the replay buffer on this step.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """

    # Act step: don't act at terminal info states or if its not our turn.
    if (not time_step.last()) and (time_step.is_simultaneous_move() or
                                   self.player_id
                                   == time_step.current_player()):
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      action, probs = self._get_action_probs(
          info_state, legal_actions, is_evaluation=is_evaluation)
    else:
      action = None
      probs = []

    # Don't mess up with the state during evaluation.
    if not is_evaluation:
      self._step_counter += 1

      if self._step_counter % self._learn_every == 0:
        self._last_loss_value = self.learn()

      if self._step_counter % self._update_target_network_every == 0:
        # state_dict method returns a dictionary containing a whole state of the
        # module.
        self.params_target_q_network = jax.tree_util.tree_map(
            lambda x: x.copy(), self.params_q_network)

      if self._prev_timestep and add_transition_record:
        # We may omit record adding here if it's done elsewhere.
        self.add_transition(self._prev_timestep, self._prev_action, time_step)

      if time_step.last():  # prepare for the next episode.
        self._prev_timestep = None
        self._prev_action = None
        return
      else:
        self._prev_timestep = time_step
        self._prev_action = action

    return rl_agent.StepOutput(action=action, probs=probs)

  def add_transition(self, prev_time_step, prev_action, time_step):
    """Adds the new transition using `time_step` to the replay buffer.

    Adds the transition from `self._prev_timestep` to `time_step` by
    `self._prev_action`.

    Args:
      prev_time_step: prev ts, an instance of rl_environment.TimeStep.
      prev_action: int, action taken at `prev_time_step`.
      time_step: current ts, an instance of rl_environment.TimeStep.
    """
    assert prev_time_step is not None
    legal_actions = (time_step.observations["legal_actions"][self.player_id])
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(
            prev_time_step.observations["info_state"][self.player_id][:]),
        action=prev_action,
        reward=time_step.rewards[self.player_id],
        next_info_state=time_step.observations["info_state"][self.player_id][:],
        is_final_step=float(time_step.last()),
        legal_actions_mask=legal_actions_mask)
    self._replay_buffer.add(transition)

  def _epsilon_greedy(self, info_state, legal_actions, epsilon):
    """Returns a valid epsilon-greedy action and valid action probs.

    Action probabilities are given by a softmax over legal q-values.

    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of legal actions at `info_state`.
      epsilon: float, probability of taking an exploratory action.

    Returns:
      A valid epsilon-greedy action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions)
    legal_one_hot = np.zeros(self._num_actions)
    legal_one_hot[legal_actions] = 1
    if np.random.rand() < epsilon:
      action = np.random.choice(legal_actions)
      probs[legal_actions] = 1.0 / len(legal_actions)
    else:
      info_state = np.reshape(info_state, [1, -1])
      q_values = self.hk_network_apply(self.params_q_network, info_state)
      legal_q_values = q_values[0] + (
          1 - legal_one_hot) * ILLEGAL_ACTION_LOGITS_PENALTY
      action = int(np.argmax(legal_q_values))
      probs[action] = 1.0
    return action, probs

  def _get_epsilon(self, is_evaluation, power=1.0):
    """Returns the evaluation or decayed epsilon value."""
    if is_evaluation:
      return 0.0
    decay_steps = min(self._step_counter, self._epsilon_decay_duration)
    decayed_epsilon = (
        self._epsilon_end + (self._epsilon_start - self._epsilon_end) *
        (1 - decay_steps / self._epsilon_decay_duration)**power)
    return decayed_epsilon

  def _loss(self, param, param_target, info_states, actions, rewards,
            next_info_states, are_final_steps, legal_actions_mask):

    q_values = self.hk_network.apply(param, info_states)
    target_q_values = self.hk_network.apply(param_target, next_info_states)
    # Sum a large negative constant to illegal action logits before taking the
    # max. This prevents illegal action values from being considered as target.
    max_next_q = jnp.max(
        target_q_values +
        (1 - legal_actions_mask) * ILLEGAL_ACTION_LOGITS_PENALTY,
        axis=-1)
    max_next_q = jax.numpy.where(
        1 - are_final_steps, max_next_q, jnp.zeros_like(max_next_q))
    target = (
        rewards + (1 - are_final_steps) * self._discount_factor * max_next_q)
    target = jax.lax.stop_gradient(target)
    predictions = jnp.sum(q_values * actions, axis=-1)
    loss_value = self.loss_func(predictions - target)
    return loss_value

  def get_update(self):

    def update(param, param_target, opt_state, info_states, actions, rewards,
               next_info_states, are_final_steps, legal_actions_mask):
      loss_val, grad_val = self._loss_and_grad(param, param_target, info_states,
                                               actions, rewards,
                                               next_info_states,
                                               are_final_steps,
                                               legal_actions_mask)
      new_param, new_opt_state = self._opt_update_fn(param, opt_state, grad_val)
      return new_param, new_opt_state, loss_val

    return update

  def _to_one_hot(self, a):
    a_one_hot = np.zeros(self._num_actions)
    a_one_hot[a] = 1.0
    return a_one_hot

  def learn(self):
    """Compute the loss on sampled transitions and perform a Q-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Returns:
      The average loss obtained on this batch of transitions or `None`.
    """

    if (len(self._replay_buffer) < self._batch_size or
        len(self._replay_buffer) < self._min_buffer_size_to_learn):
      return None

    transitions = self._replay_buffer.sample(self._batch_size)
    info_states = np.asarray([t.info_state for t in transitions])
    actions = np.asarray([self._to_one_hot(t.action) for t in transitions])
    rewards = np.asarray([t.reward for t in transitions])
    next_info_states = np.asarray([t.next_info_state for t in transitions])
    are_final_steps = np.asarray([t.is_final_step for t in transitions])
    legal_actions_mask = np.asarray([t.legal_actions_mask for t in transitions])

    self.params_q_network, self._opt_state, loss_val = self._jit_update(
        self.params_q_network, self.params_target_q_network, self._opt_state,
        info_states, actions, rewards, next_info_states, are_final_steps,
        legal_actions_mask)

    return loss_val

  @property
  def q_values(self):
    return self._q_values

  @property
  def replay_buffer(self):
    return self._replay_buffer

  @property
  def loss(self):
    return self._last_loss_value

  @property
  def prev_timestep(self):
    return self._prev_timestep

  @property
  def prev_action(self):
    return self._prev_action

  @property
  def step_counter(self):
    return self._step_counter



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements approval voting method.

Based on: https://en.wikipedia.org/wiki/Approval_voting.
"""

from open_spiel.python.voting import base


# This seems arbitrary.. is there something sensible we should default to?
DEFAULT_K = 3


class ApprovalVoting(base.AbstractVotingMethod):
  """Implements approval voting."""

  def __init__(self, k: int = 1):
    """Construct an k-Approval voting scheme.

    Note: there are no checks on the length of the votes and how they relate to 
    the value of k. So, the user is responsible for appropriately balancing the
    lengths of the votes appropriately.

    Arguments:
      k: the number of top positions to count in each vote.
    """
    self._k = k

  def name(self) -> str:
    return f"approval(k={self._k})"

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    scores = {alternative: 0 for alternative in profile.alternatives}
    for vote in profile.votes:
      vote_len = len(vote.vote)
      for i in range(self._k):
        if i >= vote_len: break
        alternative = vote.vote[i]
        scores[alternative] += vote.weight
    sorted_scores = sorted(scores.items(), key=lambda item: item[1],
                           reverse=True)
    outcome = base.RankOutcome()
    outcome.unpack_from(sorted_scores)
    return outcome



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A few prompts for performing sentiment analysis with llms.
"""

from open_spiel.python.games.chat_games.envs.payoffs import utils as payoff_utils
from open_spiel.python.games.chat_games.envs.utils import text

# maybe sentiment score should look at summarized dialogue and predict how
# happy each participant is with the state of things
MIN_PAYOFF = -10
MAX_PAYOFF = 10
PAYOFF_PROMPT = [
    'You are an assistant designed to analyze sentiment from speech data. ' +
    'Users will paste in a string of text and you will respond with an ' +
    'assessment of the response of the speaker. Rate on a scale of 1-10 ' +
    '(10 being highest). Explain why this rating was given. Here are a few ' +
    'examples.',
    '#' * 20,
    'message: Hi Bob, I am sorry you think the rent increase is unfair, but ' +
    'the economy has forced my hand. If I cannot afford it, you will have to ' +
    'leave.',
    f'sentiment_score for Bob: 3/{MAX_PAYOFF}.',
    'message: Hi Bob, you are right. The rent is expensive and it is not fair' +
    '_to raise it that high. I will compromise and only raise it by 3 percent.',
    f'sentiment_score for Bob: 7/{MAX_PAYOFF}.',
    'message: Hi Alice, I think that rent increase is really unfair. I think ' +
    'you are trying to take advantage of me and I do not appreciate it.',
    f'sentiment_score for Alice: 3/{MAX_PAYOFF}.',
    'message: Hi Alice, the rent is expensive but it is worth it and I am ' +
    'willing to pay you a higher rent.',
    f'sentiment_score for Alice: 8/{MAX_PAYOFF}.',
    '#' * 20,
    'Now provide a rating for the following message.',
    'message: {m}',
    'sentiment score for {p}: ']
PAYOFF_PROMPT = '\n\n'.join(text.wrap(PAYOFF_PROMPT))

PAYOFF_OBS_TRANS_PREFIX = ''
PAYOFF_OBS_TRANS_POSTFIX = ''

PAYOFF = payoff_utils.Payoff(PAYOFF_PROMPT,
                             MIN_PAYOFF,
                             MAX_PAYOFF,
                             PAYOFF_OBS_TRANS_PREFIX,
                             PAYOFF_OBS_TRANS_POSTFIX)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers to work with PrefLib data."""

import pyspiel
from open_spiel.python.voting import base


def parse_preflib_data(string_data: str) -> base.PreferenceProfile:
  """Parses the contents of a PrefLib data file.

  Currently only supports SOC and SOI. See https://www.preflib.org/format.

  Args:
    string_data: the name of the file to parse.

  Returns:
    A preference profile.
  """
  lines = string_data.split("\n")
  alternatives = []
  num_alternatives = None
  num_votes = None
  profile = base.PreferenceProfile()
  for raw_line in lines:
    line = raw_line.strip()
    if not line: continue
    if line.startswith("#"):
      parts = line.split(" ")
      if line.startswith("# DATA TYPE: "):
        assert(parts[3] == "soc" or parts[3] == "soi")
      elif line.startswith("# NUMBER ALTERNATIVES:"):
        num_alternatives = int(parts[3])
        alternatives = [None] * num_alternatives
      elif line.startswith("# NUMBER VOTERS:"):
        num_votes = int(parts[3])
      elif line.startswith("# ALTERNATIVE NAME "):
        num = int(parts[3].split(":")[0])
        index_of_colon = line.index(":")
        assert 1 <= num <= num_alternatives
        alternatives[num-1] = line[index_of_colon+2:]
    else:
      if profile.num_alternatives() == 0:
        profile = base.PreferenceProfile(alternatives=alternatives)
      index_of_colon = line.index(":")
      weight = int(line[:index_of_colon])
      vote_parts = line[index_of_colon+2:].split(",")
      vote = [alternatives[int(part) - 1] for part in vote_parts]
      if weight > 0:
        profile.add_vote(vote, weight)
  assert num_votes == profile.num_votes()
  return profile


def parse_preflib_datafile(filename: str) -> base.PreferenceProfile:
  """Parses a Preflib data file.

  Currently only supports SOC and SOI. See https://www.preflib.org/format.

  Args:
    filename: the name of the file to parse.

  Returns:
    A preference profile.
  """
  contents = pyspiel.read_contents_from_file(filename, "r")
  return parse_preflib_data(contents)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Borda's method.

Based on: https://en.wikipedia.org/wiki/Borda_count.
"""

from open_spiel.python.voting import base


class BordaVoting(base.AbstractVotingMethod):
  """Implements Borda's method of voting."""

  def __init__(self):
    pass

  def name(self) -> str:
    return "borda"

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    scores = {}
    for alternative in profile.alternatives:
      scores[alternative] = 0
    for vote in profile.votes:
      # Do we need a check here for the length of the vote?
      points = len(vote.vote) - 1
      for alternative in vote.vote:
        scores[alternative] += (points * vote.weight)
        points -= 1
      assert points == -1
    sorted_scores = sorted(scores.items(), key=lambda item: item[1],
                           reverse=True)
    outcome = base.RankOutcome()
    outcome.unpack_from(sorted_scores)
    return outcome



# This file will be replaced in the wheel with a hard-coded version. This only
# exists to allow running directly from source without installing (not
# recommended, but supported).

from __future__ import annotations

import re
from pathlib import Path

DIR = Path(__file__).parent.resolve()

input_file = DIR.parent / "include/pybind11/detail/common.h"
regex = re.compile(
    r"""
\#define \s+ PYBIND11_VERSION_MAJOR \s+ (?P<major>\d+) .*?
\#define \s+ PYBIND11_VERSION_MINOR \s+ (?P<minor>\d+) .*?
\#define \s+ PYBIND11_VERSION_PATCH \s+ (?P<patch>\S+)
""",
    re.MULTILINE | re.DOTALL | re.VERBOSE,
)

match = regex.search(input_file.read_text(encoding="utf-8"))
assert match, "Unable to find version in pybind11/detail/common.h"
__version__ = "{major}.{minor}.{patch}".format(**match.groupdict())


def _to_int(s: str) -> int | str:
    try:
        return int(s)
    except ValueError:
        return s


version_info = tuple(_to_int(s) for s in __version__.split("."))



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example: using MMD with dilated entropy to solve for QRE in a matrix Game."""

from absl import app
from absl import flags

from open_spiel.python.algorithms import mmd_dilated
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("iterations", 1000, "Number of iterations")
flags.DEFINE_float(
    "alpha", 0.1, "QRE parameter, larger value amounts to more regularization")
flags.DEFINE_integer("print_freq", 100, "How often to print the gap")

# create pyspiel perturbed RPS matrix game

game = pyspiel.create_matrix_game([[0, -1, 3], [1, 0, -3], [-3, 3, 0]],
                                  [[0, 1, -3], [-1, 0, 3], [3, -3, 0]])

game = pyspiel.convert_to_turn_based(game)


def main(_):
  mmd = mmd_dilated.MMDDilatedEnt(game, FLAGS.alpha)
  for i in range(FLAGS.iterations):
    mmd.update_sequences()
    if i % FLAGS.print_freq == 0:
      conv = mmd.get_gap()
      print("Iteration {} gap {}".format(i, conv))

  # Extract policies for both players
  print(mmd.get_policies().action_probability_array)
  # Note the sequence form and behavioural-form coincide
  # for a normal-form game (sequence form has extra root value of 1)
  print(mmd.current_sequences())


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.schulze."""

from absl.testing import absltest

from open_spiel.python.voting import base
from open_spiel.python.voting import schulze


class SchulzeTest(absltest.TestCase):
  def test_shulze_construction(self):
    method = schulze.SchulzeVoting()
    self.assertEqual(method.name(), "schulze")

  def test_shulze_wikipedia_example(self):
    votes = [
        base.WeightedVote(5, ["A", "C", "B", "E", "D"]),
        base.WeightedVote(5, ["A", "D", "E", "C", "B"]),
        base.WeightedVote(8, ["B", "E", "D", "A", "C"]),
        base.WeightedVote(3, ["C", "A", "B", "E", "D"]),
        base.WeightedVote(7, ["C", "A", "E", "B", "D"]),
        base.WeightedVote(2, ["C", "B", "A", "D", "E"]),
        base.WeightedVote(7, ["D", "C", "E", "B", "A"]),
        base.WeightedVote(8, ["E", "B", "A", "D", "C"])
    ]
    profile = base.PreferenceProfile(votes=votes,
                                     alternatives=["A", "B", "C", "D", "E"])
    method = schulze.SchulzeVoting()
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking, ["E", "A", "C", "B", "D"])
    self.assertListEqual(outcome.scores, [111, 88, 62, 33, 0])

  def test_meeple_pentathlon(self):
    alternatives = ["A", "B", "C"]
    votes = [
        base.WeightedVote(1, ["A", "B", "C"]),
        base.WeightedVote(1, ["A", "C", "B"]),
        base.WeightedVote(2, ["C", "A", "B"]),
        base.WeightedVote(1, ["B", "C", "A"])
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    method = schulze.SchulzeVoting()
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking, ["C", "A", "B"])
    self.assertListEqual(outcome.scores, [7, 4, 0])


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Coalitional Games in Open Spiel."""

import abc
import numpy as np


class CoalitionalGame(abc.ABC):
  """An abstract class for computing the value of a coalition."""

  def __init__(self, num_players: int):
    self._num_players = num_players

  @abc.abstractmethod
  def coalition_value(self, coalition: np.ndarray) -> float:
    """Returns the value of the coalition (the characteristic function).

    Args:
      coalition: an array of size num_players of ones (indicating player is
        included) and zeroes (the player is excluded).
    """

  def coalition_values(self, coalitions: np.ndarray) -> np.ndarray:
    """Returns the values of a batch of coalitions.

    Override to provide faster versions depending on the game.

    Args:
      coalitions: batch_size by num_players array of coalitions.
    """
    batch_size = coalitions.shape[0]
    return np.asarray(
        [self.coalition_value(coalitions[i]) for i in range(batch_size)]
    )

  def num_players(self) -> int:
    """Returns the number of players."""
    return self._num_players



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Basic tests for Elo."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.algorithms import elo
import pyspiel

SEED = 0


class EloTest(absltest.TestCase):
  """Elo rating system tests."""

  def test_simple_case_meeple_pentathlon(self):
    # Meeple Pentathlon example from the VasE paper
    # (https://arxiv.org/abs/2312.03121)
    #    1: A > B > C
    #    1: A > C > B
    #    2: C > A > B
    #    1: B > C > A
    # Here, the first and last player have provably equal Elo ratings.
    win_matrix = np.asarray([[0, 4, 2], [1, 0, 2], [3, 3, 0]])
    ratings = elo.compute_ratings_from_matrices(win_matrix)
    self.assertAlmostEqual(ratings[0], ratings[2])
    self.assertLess(ratings[1], ratings[0])
    self.assertLess(ratings[1], ratings[2])

    # Now, from match records.
    match_records = []
    match_records.extend([pyspiel.elo.MatchRecord("A", "B")] * 4)
    match_records.extend([pyspiel.elo.MatchRecord("A", "C")] * 2)
    match_records.extend([pyspiel.elo.MatchRecord("B", "A")] * 1)
    match_records.extend([pyspiel.elo.MatchRecord("B", "C")] * 2)
    match_records.extend([pyspiel.elo.MatchRecord("C", "A")] * 3)
    match_records.extend([pyspiel.elo.MatchRecord("C", "B")] * 3)
    ratings_map = pyspiel.elo.compute_ratings_from_match_records(match_records)
    self.assertAlmostEqual(ratings_map["A"], ratings[0])
    self.assertAlmostEqual(ratings_map["B"], ratings[1])
    self.assertAlmostEqual(ratings_map["C"], ratings[2])

  def test_simple_case_direct_pyspiel(self):
    # Simple case: a > b. First, specify the draws matrix. Calls the Elo
    # wrapper directly.
    ratings1 = pyspiel.elo.compute_ratings_from_matrices(
        win_matrix=[[0, 2], [1, 0]], draw_matrix=[[0, 0], [0, 0]]
    )
    self.assertGreater(ratings1[0], ratings1[1])
    self.assertAlmostEqual(ratings1[1], 0.0, places=6)

    # Testing default when excluding draws matrix.
    ratings2 = pyspiel.elo.compute_ratings_from_matrices(
        win_matrix=[[0, 2], [1, 0]]
    )
    self.assertAlmostEqual(ratings1[0], ratings2[0], places=6)
    self.assertAlmostEqual(ratings1[1], ratings2[1], places=6)


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Basic tests for Soft Condorcet Optimization."""

from absl.testing import absltest
import numpy as np
import pyspiel
from open_spiel.python.voting import base
from open_spiel.python.voting import soft_condorcet_optimization as sco

SEED = 0


class SCOTest(absltest.TestCase):
  """Soft Condorcet Optimization tests."""

  def test_simple_case(self):
    # Simple case: a > b > c
    profile = base.PreferenceProfile(votes=[["a", "b", "c"]])
    solver = sco.SoftCondorcetOptimizer(profile, temperature=1)
    ratings, ranking = solver.run_solver(1000, learning_rate=0.01)
    alt_idx = profile.alternatives_dict
    for alt in ranking:
      print(f"{alt}: {ratings[alt_idx[alt]]}")
    self.assertGreater(ratings[0], ratings[1])
    self.assertGreater(ratings[1], ratings[2])

  def test_meeple_pentathlon(self):
    """Meeple pentathlon from the VasE paper."""
    profile = base.PreferenceProfile(
        votes=[
            ["A", "B", "C"],
            ["A", "C", "B"],
            ["C", "A", "B"],
            ["C", "A", "B"],
            ["B", "C", "A"],
        ]
    )
    solver = sco.SoftCondorcetOptimizer(profile, batch_size=4, temperature=1)
    ratings, ranking = solver.run_solver(10000, learning_rate=0.01)
    alt_idx = profile.alternatives_dict
    for alt in ranking:
      print(f"{alt}: {ratings[alt_idx[alt]]}")
    # Correct ranking is C > A > B.
    self.assertGreater(ratings[2], ratings[0])
    self.assertGreater(ratings[0], ratings[1])

  def test_cpp_meeple_pentathlon_sigmoid(self):
    # Tests the C++ implementation of the SCO with sigmoid solver.
    profile = base.PreferenceProfile(
        votes=[
            ["A", "B", "C"],
            ["A", "C", "B"],
            ["C", "A", "B"],
            ["C", "A", "B"],
            ["B", "C", "A"],
        ]
    )
    cpp_sco_solver = pyspiel.sco.SoftCondorcetOptimizer(
        profile.to_list_of_tuples(),
        rating_lower_bound=-100.0,
        rating_upper_bound=100.0,
        batch_size=4,
        temperature=1,
        rng_seed=SEED,
    )
    cpp_sco_solver.run_solver(10000, learning_rate=0.01)
    ratings_dict = cpp_sco_solver.ratings()
    self.assertGreater(ratings_dict["C"], ratings_dict["A"])
    self.assertGreater(ratings_dict["A"], ratings_dict["B"])

  def test_cpp_meeple_pentathlon_fenchel_young(self):
    # Tests the C++ implementation of the FY solver.
    profile = base.PreferenceProfile(
        votes=[
            ["A", "B", "C"],
            ["A", "C", "B"],
            ["C", "A", "B"],
            ["C", "A", "B"],
            ["B", "C", "A"],
        ]
    )
    cpp_fy_solver = pyspiel.sco.FenchelYoungOptimizer(
        profile.to_list_of_tuples(),
        rating_lower_bound=-100.0,
        rating_upper_bound=100.0,
        batch_size=4,
        rng_seed=SEED,
    )
    cpp_fy_solver.run_solver(10000, learning_rate=0.01)
    ratings_dict = cpp_fy_solver.ratings()
    # C is not necessarily better than A here, just like with Elo.
    # But both should have higher ratings than B.
    self.assertGreater(ratings_dict["C"], ratings_dict["B"])
    self.assertGreater(ratings_dict["A"], ratings_dict["B"])


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Methods to compute the core based on linear programming.

This file contains methods to compute the core using LPs referred to in
Yan & Procaccia '21: https://ojs.aaai.org/index.php/AAAI/article/view/16721
"""

import itertools

from typing import Any, Callable, List, Tuple

import cvxpy as cp
import numpy as np

from open_spiel.python.coalitional_games import coalitional_game


ConstraintsSamplingFuncType = Callable[
    [coalitional_game.CoalitionalGame, cp.Variable, cp.Variable, List[Any]],
    Any,
]


def add_all_constraints(
    game: coalitional_game.CoalitionalGame,
    x: cp.Variable,
    e: cp.Variable,
    constraints: List[Any]):
  # \sum x_i + e >= v(S), for all subsets S \subseteq N
  for c in itertools.product([0, 1], repeat=game.num_players()):
    coalition = np.asarray(c)
    val_coalition = game.coalition_value(coalition)
    constraints.append(x @ coalition + e >= val_coalition)


def make_uniform_sampling_constraints_function(
    num: int,
) -> ConstraintsSamplingFuncType:
  """Simple uniform constraint sampler (with replacement)."""

  def func(game: coalitional_game.CoalitionalGame,
           x: cp.Variable, e: cp.Variable, constraints: List[Any]):
    for _ in range(num):
      coalition = np.random.randint(2, size=game.num_players())
      val_coalition = game.coalition_value(coalition)
      constraints.append(x @ coalition + e >= val_coalition)
  return func


def solve_least_core_lp(
    game: coalitional_game.CoalitionalGame,
    constraint_function: ConstraintsSamplingFuncType,
) -> Tuple[np.ndarray, float]:
  """Solve the LP described in Yan & Procaccia, equation (1).

  This LP enumerates all (exponentially many!) possible coalitions, with one
  constraint per coalition. Will not scale to games with too many players.

  Args:
    game: the game the LP solves.
    constraint_function: function that adds the constraints

  Returns:
    solution: an array with num_players entries,
    epsilon: the lowest epsilon.
  """
  # TODO(author5): handle errors gracefully. E.g. if solving the LP fails.

  num_players = game.num_players()
  val_gc = game.coalition_value(np.ones(game.num_players()))

  # min e
  # indices 0 - n-1 correspond to x_i, index n corresponds to e
  x = cp.Variable(num_players, nonneg=True)
  e = cp.Variable()  # note: epsilon can be negative when the core is non-empty!

  objective = cp.Minimize(e)
  constraints = []

  # \sum_{i in N} x_i  = v(N)
  constraints.append(x @ np.ones(num_players) == val_gc)

  # Add the constraints
  constraint_function(game, x, e, constraints)

  prob = cp.Problem(objective, constraints)
  _ = prob.solve(solver=cp.SCS, eps=1e-6)
  # The optimal value for x is stored in `x.value`.

  return x.value, e.value



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Regret-Matching Algorithm.

This is an N-player implementation of the regret-matching algorithm described in
Hart & Mas-Colell 2000:
https://onlinelibrary.wiley.com/doi/abs/10.1111/1468-0262.00153
"""

import numpy as np
from open_spiel.python.algorithms import nfg_utils


# Start with initial regrets of 1 / denom
INITIAL_REGRET_DENOM = 1e6


def _partial_multi_dot(player_payoff_tensor, strategies, index_avoided):
  """Computes a generalized dot product avoiding one dimension.

  This is used to directly get the expected return of a given action, given
  other players' strategies, for the player indexed by index_avoided.
  Note that the numpy.dot function is used to compute this product, as it ended
  up being (Slightly) faster in performance tests than np.tensordot. Using the
  reduce function proved slower for both np.dot and np.tensordot.

  Args:
    player_payoff_tensor: payoff tensor for player[index_avoided], of dimension
      (dim(vector[0]), dim(vector[1]), ..., dim(vector[-1])).
    strategies: Meta strategy probabilities for each player.
    index_avoided: Player for which we do not compute the dot product.

  Returns:
    Vector of expected returns for each action of player [the player indexed by
      index_avoided].
  """
  new_axis_order = [index_avoided] + [
      i for i in range(len(strategies)) if (i != index_avoided)
  ]
  accumulator = np.transpose(player_payoff_tensor, new_axis_order)
  for i in range(len(strategies) - 1, -1, -1):
    if i != index_avoided:
      accumulator = np.dot(accumulator, strategies[i])
  return accumulator


def _regret_matching_step(payoff_tensors, strategies, regrets, gamma):
  """Does one step of the projected replicator dynamics algorithm.

  Args:
    payoff_tensors: List of payoff tensors for each player.
    strategies: List of the strategies used by each player.
    regrets: List of cumulative regrets used by each player.
    gamma: Minimum exploratory probability term.

  Returns:
    A list of updated strategies for each player.
  """

  # TODO(author4): Investigate whether this update could be fully vectorized.
  new_strategies = []
  for player in range(len(payoff_tensors)):
    current_payoff_tensor = payoff_tensors[player]
    current_strategy = strategies[player]

    values_per_strategy = _partial_multi_dot(current_payoff_tensor, strategies,
                                             player)
    average_return = np.dot(values_per_strategy, current_strategy)
    regrets[player] += values_per_strategy - average_return

    updated_strategy = regrets[player].copy()
    updated_strategy[updated_strategy < 0] = 0.0
    sum_regret = updated_strategy.sum()
    uniform_strategy = np.ones(len(updated_strategy)) / len(updated_strategy)

    if sum_regret > 0:
      updated_strategy /= sum_regret
      updated_strategy = gamma * uniform_strategy + (1 -
                                                     gamma) * updated_strategy
    else:
      updated_strategy = uniform_strategy

    new_strategies.append(updated_strategy)
  return new_strategies


def regret_matching(payoff_tensors,
                    initial_strategies=None,
                    iterations=int(1e5),
                    gamma=1e-6,
                    average_over_last_n_strategies=None,
                    **unused_kwargs):
  """Runs regret-matching for the stated number of iterations.

  Args:
    payoff_tensors: List of payoff tensors for each player.
    initial_strategies: Initial list of the strategies used by each player, if
      any. Could be used to speed up the search by providing a good initial
      solution.
    iterations: Number of algorithmic steps to take before returning an answer.
    gamma: Minimum exploratory probability term.
    average_over_last_n_strategies: Running average window size for average
      policy computation. If None, use the whole trajectory.
    **unused_kwargs: Convenient way of exposing an API compatible with other
      methods with possibly different arguments.

  Returns:
    RM-computed strategies.
  """
  number_players = len(payoff_tensors)
  # Number of actions available to each player.
  action_space_shapes = payoff_tensors[0].shape

  # If no initial starting position is given, start with uniform probabilities.
  new_strategies = initial_strategies or [
      np.ones(action_space_shapes[k]) / action_space_shapes[k]
      for k in range(number_players)
  ]

  regrets = [
      np.ones(action_space_shapes[k]) / INITIAL_REGRET_DENOM
      for k in range(number_players)
  ]

  averager = nfg_utils.StrategyAverager(number_players, action_space_shapes,
                                        average_over_last_n_strategies)
  averager.append(new_strategies)

  for _ in range(iterations):
    new_strategies = _regret_matching_step(payoff_tensors, new_strategies,
                                           regrets, gamma)
    averager.append(new_strategies)
  return averager.average_strategies()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A base environment for scheduling with private info.
"""

import dataclasses

from open_spiel.python.games.chat_games.envs.comm_substrates import schedules
from open_spiel.python.games.chat_games.envs.utils import header


action_keys = tuple(['tone'])
info_keys = tuple(['day_prefs', 'ooo_days'])

w_opts = (schedules.W_OPTS_PREFIX +
          'Out of Office Days:\n{ooo_days}\n\n' +
          'Day Preferences:\n{day_prefs}\n\n' +
          'Tone: Use a {tone} tone.\n' +
          schedules.PLAIN)

ab = schedules.PLAIN.format(sender='Alice', receiver='Bob') + '\n\n'
ba = schedules.PLAIN.format(sender='Bob', receiver='Alice') + '\n\n'
cd = schedules.PLAIN.format(sender='Charlie', receiver='David') + '\n\n'
dc = schedules.PLAIN.format(sender='David', receiver='Charlie') + '\n\n'

context = '''An intelligent assistant is looking at dialogs between two people
trying to decide when to meet, and determines whether they have managed to agree
on a meeting time, and if so when the meeting is set to occur.

Example 1:
{s1}Hi Bob, can we meet on Monday?
{s2}No, I am out of the office on Monday. How about Tuesday?
{s3}Well, I am in the office on Tuesday but I would rather keep my schedule
free. Can we do Friday instead.
{s4}Great, Friday it is. See you then!

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
<END OF EXAMPLE 1>
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Summary: Alice suggests Monday, Bob declines. Bob suggests Tuesday. Alice
declines. Alice suggests Friday. Bob agrees.
Outcome Summary: Meeting agreed on Friday.

Example 2:
{s5}Hi David, would you like to meet on Friday?
{s6}I hate working on Fridays. Can't we meet on Tuesday?
{s7}On Tuesday I am out of the office, and Wednesday also doesn't work for me.
How do you feel about meeting on Saturday?
{s8}Excellent, let's meet on Saturday.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
<END OF EXAMPLE 2>
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Summary: Charlie suggests Friday. David declines. David suggests Tuesday.
Charlie declines. Charlie suggests Saturday. David agrees.
Outcome Summary: Meeting agreed on Saturday.

Example 3:
'''.format(s1=ab, s2=ba, s3=ab, s4=ba, s5=cd, s6=dc, s7=cd, s8=dc)

HEADER = header.Header(schedules.PLAIN,
                       w_opts,
                       schedules.strip_msg,
                       schedules.SPECIAL_CHARS,
                       action_keys,
                       info_keys,
                       context)


@dataclasses.dataclass(frozen=True)
class Scenario(header.BaseScenario):
  ooo_days: str
  day_prefs: str
  tone: str = 'calm'



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Crowd Modelling game."""

from absl.testing import absltest
from open_spiel.python.mfg.games import normal_form_game
import pyspiel

MFG_STR_CONST = "_a"


class MFGNormalFormGameTest(absltest.TestCase):

  def test_load(self):
    game = pyspiel.load_game("mean_field_nfg")
    game.new_initial_state()

  def test_create(self):
    """Checks we can create the game and clone states."""
    game = normal_form_game.MFGNormalFormGame()
    self.assertEqual(
        game.get_type().dynamics, pyspiel.GameType.Dynamics.MEAN_FIELD
    )
    print("Num distinct actions:", game.num_distinct_actions())
    state = game.new_initial_state()
    clone = state.clone()
    print("Initial state:", state)
    print("Cloned initial state:", clone)

  def test_create_with_params(self):
    game = pyspiel.load_game("mean_field_nfg(num_actions=10)")
    self.assertEqual(game.num_actions, 10)

  def test_reward(self):
    game = normal_form_game.MFGNormalFormGame()
    state = game.new_initial_state()
    self.assertEqual(state.current_player(), pyspiel.PlayerId.DEFAULT_PLAYER_ID)

    state.apply_action(0)
    self.assertEqual(state.current_player(), pyspiel.PlayerId.MEAN_FIELD)
    state.update_distribution([1.0, 0.0, 0.0])
    self.assertAlmostEqual(state.rewards()[0], 10.0)
    self.assertAlmostEqual(state.returns()[0], 10.0)

    state = game.new_initial_state()
    state.apply_action(0)
    state.update_distribution([0.0, 1.0, 0.0])
    self.assertAlmostEqual(state.rewards()[0], -20.0)
    self.assertAlmostEqual(state.returns()[0], -20.0)

    self.assertTrue(state.is_terminal())


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Latin American Python Dominoes."""


from absl.testing import absltest
from open_spiel.python.games import team_dominoes
import pyspiel


class DominoesTest(absltest.TestCase):

  def test_game_from_cc(self):
    """Runs our standard game tests, checking API consistency."""
    game = pyspiel.load_game("python_team_dominoes")
    pyspiel.random_sim_test(game, num_sims=100, serialize=False, verbose=True)

  def test_single_deterministic_game_1(self):
    """Runs a single game where tiles and actions chose deterministically."""
    game = pyspiel.load_game("python_team_dominoes")
    state = game.new_initial_state()
    hand0 = [
        (1.0, 3.0),
        (0.0, 5.0),
        (1.0, 1.0),
        (2.0, 3.0),
        (4.0, 5.0),
        (3.0, 5.0),
        (0.0, 1.0),
    ]
    hand1 = [
        (2.0, 5.0),
        (3.0, 4.0),
        (2.0, 2.0),
        (0.0, 4.0),
        (3.0, 3.0),
        (2.0, 6.0),
        (1.0, 6.0),
    ]
    hand2 = [
        (5.0, 6.0),
        (6.0, 6.0),
        (1.0, 4.0),
        (2.0, 4.0),
        (4.0, 4.0),
        (0.0, 0.0),
        (1.0, 5.0),
    ]
    hand3 = [
        (4.0, 6.0),
        (0.0, 2.0),
        (0.0, 3.0),
        (3.0, 6.0),
        (5.0, 5.0),
        (1.0, 2.0),
        (0.0, 6.0),
    ]

    self.deal_hands(state, [hand0, hand1, hand2, hand3])

    self.apply_action(state, team_dominoes.Action(0, (3.0, 4.0), None))
    self.apply_action(state, team_dominoes.Action(1, (2.0, 4.0), 4.0))
    self.apply_action(state, team_dominoes.Action(2, (1.0, 2.0), 2.0))
    self.apply_action(state, team_dominoes.Action(3, (0.0, 3.0), 3.0))

    self.apply_action(state, team_dominoes.Action(0, (1.0, 3.0), 1.0))
    self.apply_action(state, team_dominoes.Action(1, (3.0, 5.0), 3.0))
    self.apply_action(state, team_dominoes.Action(2, (0.0, 2.0), 0.0))
    self.apply_action(state, team_dominoes.Action(3, (2.0, 5.0), 2.0))

    self.apply_action(state, team_dominoes.Action(0, (1.0, 5.0), 5.0))
    self.apply_action(state, team_dominoes.Action(1, (0.0, 5.0), 5.0))
    self.apply_action(state, team_dominoes.Action(2, (1.0, 1.0), 1.0))
    self.apply_action(state, team_dominoes.Action(3, (0.0, 6.0), 0.0))

    self.apply_action(state, team_dominoes.Action(0, (3.0, 6.0), 6.0))
    self.apply_action(state, team_dominoes.Action(1, (1.0, 6.0), 1.0))
    self.apply_action(state, team_dominoes.Action(2, (5.0, 6.0), 6.0))
    self.apply_action(state, team_dominoes.Action(3, (3.0, 3.0), 3.0))

    self.apply_action(state, team_dominoes.Action(0, (4.0, 5.0), 5.0))
    self.apply_action(state, team_dominoes.Action(1, (4.0, 6.0), 4.0))
    self.apply_action(state, team_dominoes.Action(3, (6.0, 6.0), 6.0))

    self.apply_action(state, team_dominoes.Action(0, (2.0, 6.0), 6.0))
    self.apply_action(state, team_dominoes.Action(1, (2.0, 2.0), 2.0))
    self.apply_action(state, team_dominoes.Action(3, (2.0, 3.0), 3.0))
    # Game is stuck! No player can play any tile as all 2.0s are played

    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns()[0], -18)
    self.assertEqual(state.returns()[1], 18)
    self.assertEqual(state.returns()[2], -18)
    self.assertEqual(state.returns()[3], 18)

  def test_single_deterministic_game_2(self):
    """Runs a single game where tiles and actions chose deterministically."""
    game = pyspiel.load_game("python_team_dominoes")
    state = game.new_initial_state()
    hand0 = [
        (0.0, 6.0),
        (3.0, 6.0),
        (1.0, 3.0),
        (1.0, 4.0),
        (5.0, 5.0),
        (0.0, 0.0),
        (2.0, 6.0),
    ]
    hand1 = [
        (1.0, 5.0),
        (2.0, 2.0),
        (0.0, 2.0),
        (0.0, 3.0),
        (4.0, 5.0),
        (6.0, 6.0),
        (5.0, 6.0),
    ]
    hand2 = [
        (2.0, 4.0),
        (3.0, 4.0),
        (3.0, 3.0),
        (0.0, 4.0),
        (1.0, 1.0),
        (1.0, 6.0),
        (3.0, 5.0),
    ]
    hand3 = [
        (0.0, 5.0),
        (0.0, 1.0),
        (4.0, 4.0),
        (2.0, 3.0),
        (1.0, 2.0),
        (2.0, 5.0),
        (4.0, 6.0),
    ]

    self.deal_hands(state, [hand0, hand1, hand2, hand3])

    self.apply_action(state, team_dominoes.Action(0, (0.0, 6.0), None))
    self.apply_action(state, team_dominoes.Action(1, (0.0, 5.0), 0.0))
    self.apply_action(state, team_dominoes.Action(2, (2.0, 6.0), 6.0))
    self.apply_action(state, team_dominoes.Action(3, (1.0, 5.0), 5.0))

    self.apply_action(state, team_dominoes.Action(0, (2.0, 3.0), 2.0))
    self.apply_action(state, team_dominoes.Action(1, (3.0, 6.0), 3.0))
    self.apply_action(state, team_dominoes.Action(2, (1.0, 3.0), 1.0))
    self.apply_action(state, team_dominoes.Action(3, (1.0, 6.0), 6.0))

    self.apply_action(state, team_dominoes.Action(0, (3.0, 5.0), 3.0))
    self.apply_action(state, team_dominoes.Action(1, (5.0, 6.0), 5.0))
    self.apply_action(state, team_dominoes.Action(2, (1.0, 1.0), 1.0))
    self.apply_action(state, team_dominoes.Action(3, (4.0, 6.0), 6.0))

    # skipped player 0 (has no 4.0 or 1.0 to play)
    self.apply_action(state, team_dominoes.Action(1, (0.0, 4.0), 4.0))
    self.apply_action(state, team_dominoes.Action(2, (0.0, 1.0), 1.0))
    # skipped player 3 (has no 0.0s to play)

    # skipped over player 0 (has no 0.0s to play)
    self.apply_action(state, team_dominoes.Action(1, (0.0, 0.0), 0.0))
    self.apply_action(state, team_dominoes.Action(2, (0.0, 3.0), 0.0))
    self.apply_action(state, team_dominoes.Action(3, (3.0, 4.0), 3.0))

    # skipped over player 0 (has no 0.0s nor 4.0s to play)
    self.apply_action(state, team_dominoes.Action(1, (0.0, 2.0), 0.0))
    self.apply_action(state, team_dominoes.Action(2, (2.0, 4.0), 2.0))
    self.apply_action(state, team_dominoes.Action(3, (1.0, 4.0), 4.0))

    # skipped over player 0 (has no 1.0s nor 4.0s to play)
    self.apply_action(state, team_dominoes.Action(1, (1.0, 2.0), 1.0))
    # player 1 won (no more tiles to play)

    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns()[0], -39)
    self.assertEqual(state.returns()[1], 39)
    self.assertEqual(state.returns()[2], -39)
    self.assertEqual(state.returns()[3], 39)

  def apply_action(self, state, action):
    actions_str = team_dominoes._ACTIONS_STR
    state.apply_action(actions_str.index(str(action)))

  def deal_hands(self, state, hands):
    deck = team_dominoes._DECK
    for hand in hands:
      for t in hand:
        state.apply_action(deck.index(t))


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python3

# 2017, Georg Sauthoff <mail@gms.tf>, GPLv3

import sys

def skip_comments(lines):
  state = 0
  for line in lines:
    n = len(line)
    l = ''
    p = 0
    while p < n:
      if state == 0:
        a = line.find('//', p)
        b = line.find('/*', p)
        if a > -1 and (a < b or b == -1):
          l += line[p:a]
          p = n
        elif b > -1 and (b < a or a == -1):
          l += line[p:b]
          p = b+2
          state = 1
        else:
          l += line[p:]
          p = n
      elif state == 1:
        a = line.rfind('*/', p)
        if a == -1:
          p = n
        else:
          p = a + 2
          state = 0
    yield l

def cond_lines(lines):
  state = 0
  pcnt = 0
  for nr, line in enumerate(lines, 1):
    if not line:
      continue
    n = len(line)
    p = 0
    do_yield = False
    while p < n:
      if state == 0:
        p = line.find('if', p)
        if p == -1:
          p = n
          continue
        if (p == 0 or not line[p-1].isalpha()) \
            and (p+2 == len(line) or not line[p+2].isalpha()):
          do_yield = True
          state = 1
        p += 2
      elif state == 1:
        do_yield = True
        p = line.find('(', p)
        if p == -1:
          p = n
        else:
          p += 1
          state = 2
          pcnt = 1
      elif state == 2:
        do_yield = True
        for p in range(p, n):
          if line[p] == '(':
            pcnt += 1
          elif line[p] == ')':
            pcnt -= 1
          if not pcnt:
            state = 0
            break
        p += 1
    if do_yield:
      yield nr

def cond_lines_from_file(filename):
  with open(filename) as f:
    yield from cond_lines(skip_comments(f))

def filter_lcov_trace(lines):
  nrs = set()
  for line in lines:
    if line.startswith('SF:'):
      nrs = set(cond_lines_from_file(line[3:-1]))
    elif line.startswith('BRDA:'):
      xs = line[5:].split(',')
      nr = int(xs[0]) if xs else 0
      if nr not in nrs:
        continue
    yield line

def filter_lcov_trace_file(s_filename, d_file):
  with open(s_filename) as f:
    for l in filter_lcov_trace(f):
      print(l, end='', file=d_file)

if __name__ == '__main__':
  #for l in cond_lines_from_file(sys.argv[1]):
  #  print(l)

  filter_lcov_trace_file(sys.argv[1], sys.stdout)

  #with open(sys.argv[1]) as f:
  #  for l in skip_comments(f):
  #    print(l)




# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
import numpy as np
from open_spiel.python.coalitional_games import least_core_lp
from open_spiel.python.coalitional_games import shapley_values
from open_spiel.python.coalitional_games import wvg


SEED = 2093777


class WeightedVotingGamesTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(SEED)

  def test_basic_wvg_equal_weights(self):
    # Equal weights.
    game = wvg.WeightedVotingGame(weights=np.asarray([10]*4), quota=35.0)
    svals = shapley_values.compute_shapley_values(game)
    self.assertTrue(np.allclose(svals, np.asarray([0.25, 0.25, 0.25, 0.25])))
    lc_imputation, epsilon = least_core_lp.solve_least_core_lp(
        game, least_core_lp.add_all_constraints)
    self.assertTrue(np.allclose(lc_imputation,
                                np.asarray([0.25, 0.25, 0.25, 0.25])))
    self.assertAlmostEqual(epsilon, 0.0)

  def test_basic_wvg_unequal_weights(self):
    # Example 2.3 of the CACGT book by by Chalkiadakis, Elkind, and Wooldridge.
    game = wvg.WeightedVotingGame(weights=np.asarray([40.0, 22.0, 30.0, 9.0]),
                                  quota=51.0)
    svals = shapley_values.compute_shapley_values(game)
    self.assertTrue(np.allclose(svals, np.asarray([1.0/3, 1.0/3, 1.0/3, 0])))
    lc_imputation, epsilon = least_core_lp.solve_least_core_lp(
        game, least_core_lp.add_all_constraints)
    print(lc_imputation)   # prints [0.33, 0.33, 0.33, 0]
    print(epsilon)         # prints 0.33
    np.testing.assert_array_almost_equal(
        lc_imputation,
        np.asarray([1.0 / 3, 1.0 / 3, 1.0 / 3, 0]),
        decimal=4,
    )
    self.assertAlmostEqual(epsilon, 1.0/3.0, delta=1e-4)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Approximate the limiting logit equilbrium (Nash) of a large normal-form game.

This is a python implementation of the Nash solver for normal-form games,
Average Deviation Incentive Descent with Adaptive Sampling (ADIDAS), from
"Sample-based Approximation of Nash in Large Many-player Games via Gradient
Descent" [Gemp et al, AAMAS 2022].

Link to paper: https://arxiv.org/abs/2106.01285.

The limiting logit equilibrium (LLE) was originally defined in "Quantal Response
Equilibria for Normal Form Games" [McKelvey & Palfrey, Games and Economic
Behavior 1995]. The LLE is a Nash equilibrium that is uniquely defined for
*almost* all games.
"""

import itertools

import time

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import misc
from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.nonsymmetric import exploitability as nonsym_exp
from open_spiel.python.algorithms.adidas_utils.helpers.nonsymmetric import game_runner as nonsym_game_runner
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability as sym_exp
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import game_runner as sym_game_runner


class ADIDAS(object):
  """Average Deviation Incentive Descent with Adaptive Sampling.

  Approximate the limiting logit equilibrium of a normal-form game.

  Attributes:
    experiment_seed: int, seed for random number generator
    random: numpy.random.RandomState object
    results: dictionary of results populated upon completion of solver
  """

  def __init__(self, seed=0):
    self.experiment_seed = seed
    self.random = np.random.RandomState(self.experiment_seed)

    self.results = None

  def estimate_exploitability_sym(self, dist, num_eval_samples, num_ckpts,
                                  num_players, game, policies):
    """Estimate exploitability via monte carlo.

    Args:
      dist: 1-d np.array, estimate of nash distribution
      num_eval_samples: int, number of samples to estimate exploitability
      num_ckpts: int, number of checkpoints (actions, policies, ...)
      num_players: int, number of players
      game: game with minimal functionality (see games/small.py)
      policies: list mapping checkpoints to policies
    Returns:
      list of exploitabilities computed using [index] monte carlo samples
    """
    pg_mean = np.zeros_like(dist)
    exps_estimated = []
    for s in range(num_eval_samples):
      base_profile = tuple([
          self.random.choice(num_ckpts, p=dist) for _ in range(num_players)
      ])
      game_queries = sym_game_runner.construct_game_queries_for_exp(
          base_profile, num_ckpts)
      game_results = sym_game_runner.run_games_and_record_payoffs(
          game_queries, game.get_payoffs_for_strategies, policies)
      pg_s = np.zeros_like(dist)
      for query, payoffs in game_results.items():
        pg_s[query[0]] = payoffs[0]
      pg_mean = (pg_mean * float(s) + pg_s) / float(s + 1)
      exps_estimated.append(pg_mean.max() - pg_mean.dot(dist))

    return exps_estimated

  def estimate_exploitability_nonsym(self, dist, num_eval_samples, num_ckpts,
                                     num_players, game, policies):
    """Estimate exploitability via monte carlo.

    Args:
      dist: list of 1-d np.arrays, estimate of nash distribution
      num_eval_samples: int, number of samples to estimate exploitability
      num_ckpts: int, number of checkpoints (actions, policies, ...)
      num_players: int, number of players
      game: game with minimal functionality (see games/small.py)
      policies: list mapping checkpoints to policies
    Returns:
      list of exploitabilities computed using [index] monte carlo samples
    """
    pg_mean = [np.zeros_like(dist_i) for dist_i in dist]
    exps_estimated = []
    for s in range(num_eval_samples):
      base_profile = tuple([
          self.random.choice(num_ckpts[i], p=dist[i])
          for i in range(num_players)
      ])
      game_queries = nonsym_game_runner.construct_game_queries_for_exp(
          base_profile, num_ckpts)
      game_results = nonsym_game_runner.run_games_and_record_payoffs(
          game_queries, game.get_payoffs_for_strategies, policies)
      for pi_query, payoffs in game_results.items():
        pi, query = pi_query
        ai = query[pi]
        pg_mean[pi][ai] += (payoffs[pi] - pg_mean[pi][ai]) / float(s + 1)
      exp_is = []
      for i in range(num_players):
        exp_is.append(pg_mean[i].max() - pg_mean[i].dot(dist[i]))
      exps_estimated.append(np.mean(exp_is))

    return exps_estimated

  def update_payoff_matrices(self, payoff_matrices, payoff_matrices_new, s):
    """Update mean of payoff matrices.

    Args:
      payoff_matrices: dictionary with keys as tuples of agents (i, j) and
          values of (2 x A x A) np.arrays, payoffs for each joint action. keys
          are sorted and arrays should be indexed in the same order
          **current mean
      payoff_matrices_new: **new sample
      s: int, sample number
    Returns:
      payoff_matrices with updated means
    """
    if payoff_matrices:
      for key in payoff_matrices_new:
        new = payoff_matrices_new[key]
        old = payoff_matrices[key]
        payoff_matrices[key] += (new - old) / float(s + 1)
    else:
      payoff_matrices = payoff_matrices_new

    return payoff_matrices

  def construct_payoff_matrices_from_samples_sym(
      self, game, dist, num_samples, policies, num_players, num_ckpts):
    """Construct payoff matrices (approx. sym. polymatrix game) from samples.

    Args:
      game: game with minimal functionality (see games/small.py)
      dist: 1-d np.array, estimate of nash distribution
      num_samples: int, `minibatch' size for stochastic gradient
      policies: list mapping checkpoints to policies
      num_players: int, number of players
      num_ckpts: int, number of checkpoints (actions, policies, ...)
    Returns:
      payoff_matrices (2 x num_ckpts x num_ckpts array) to compute adidas grad
    """
    payoff_matrices = np.zeros((2, num_ckpts, num_ckpts))
    for _ in range(num_samples):
      base_profile = tuple([
          self.random.choice(num_ckpts, p=dist) for _ in range(num_players)
      ])
      game_queries = sym_game_runner.construct_game_queries(
          base_profile, num_ckpts)
      game_results = sym_game_runner.run_games_and_record_payoffs(
          game_queries, game.get_payoffs_for_strategies, policies)
      payoff_matrices += sym_game_runner.form_payoff_matrices(
          game_results, num_ckpts) / float(num_samples)
    return payoff_matrices

  def construct_payoff_matrices_exactly_sym(
      self, game, dist, num_players):
    """Construct payoff matrices exactly (expected sym. polymatrix game).

    Args:
      game: game with minimal functionality (see games/small.py)
      dist: 1-d np.array, estimate of nash distribution
      num_players: int, number of players
    Returns:
      payoff_matrices (2 x A x A array) to compute adidas gradient
    """
    sym_nash = [dist for _ in range(num_players)]
    pt = game.payoff_tensor()
    payoff_matrix_exp_0 = misc.pt_reduce(pt[0], sym_nash, [0, 1])
    payoff_matrix_exp_1 = misc.pt_reduce(pt[1], sym_nash, [0, 1])
    payoff_matrices = np.stack((payoff_matrix_exp_0, payoff_matrix_exp_1))
    return payoff_matrices

  def construct_payoff_matrices_from_samples_nonsym(
      self, game, dist, num_samples, policies, num_players, num_ckpts):
    """Construct payoff matrices (approx. nonsym. polymatrix) from samples.

    Args:
      game: game with minimal functionality (see games/small.py)
      dist: list of 1-d np.arrays, estimate of nash distribution
      num_samples: int, `minibatch' size for stochastic gradient
      policies: list mapping checkpoints to policies
      num_players: int, number of players
      num_ckpts: int, number of checkpoints (actions, policies, ...)
    Returns:
      payoff_matrices: dictionary with keys as tuples of agents (i, j) and
          values of (2 x A x A) np.arrays, payoffs for each joint action. keys
          are sorted and arrays should be indexed in the same order
    """
    payoff_matrices = None
    for s in range(num_samples):
      base_profile = tuple([
          self.random.choice(num_ckpts[i], p=dist[i])
          for i in range(num_players)
      ])
      game_queries = nonsym_game_runner.construct_game_queries(
          base_profile, num_ckpts)
      game_results = nonsym_game_runner.run_games_and_record_payoffs(
          game_queries, game.get_payoffs_for_strategies, policies)
      payoff_matrices_new = nonsym_game_runner.form_payoff_matrices(
          game_results, num_ckpts)
      payoff_matrices = self.update_payoff_matrices(payoff_matrices,
                                                    payoff_matrices_new,
                                                    s)
    return payoff_matrices

  def construct_payoff_matrices_exactly_nonsym(
      self, game, dist, num_players):
    """Construct payoff matrices exactly (expected nonsym. polymatrix game).

    Args:
      game: game with minimal functionality (see games/small.py)
      dist: list of 1-d np.arrays, estimate of nash distribution
      num_players: int, number of players
    Returns:
      payoff_matrices: dictionary with keys as tuples of agents (i, j) and
          values of (2 x A x A) np.arrays, payoffs for each joint action. keys
          are sorted and arrays should be indexed in the same order
    """
    pt = game.payoff_tensor()
    payoff_matrices = {}
    for pi, pj in itertools.combinations(range(num_players), 2):
      key = (pi, pj)
      pt_i = misc.pt_reduce(pt[pi], dist, [pi, pj])
      pt_j = misc.pt_reduce(pt[pj], dist, [pi, pj])
      payoff_matrices[key] = np.stack((pt_i, pt_j), axis=0)
    return payoff_matrices

  def approximate_nash(self, game, solver, sym,
                       num_iterations=10000, num_samples=1,
                       num_eval_samples=int(10e4), approx_eval=False,
                       exact_eval=False, avg_trajectory=False,
                       return_trajectory=False):
    """Runs solver on game.

    Args:
      game: game with minimal functionality (see games/small.py)
      solver: gradient solver (see utils/updates.py)
      sym: bool, true if the game is symmetric across players
      num_iterations: int, number of incremental updates
      num_samples: int, `minibatch' size for stochastic gradient
      num_eval_samples: int, number of samples to estimate exploitability
        default = # of samples for P[|sample_payoff-true| > C/100] < ~5e-7%
        where C = pt.max() - pt.min();
        P[|pt_grad|_inf <= C/100] > (1-5e-7)^num_actions
      approx_eval: bool, whether to evaluate exploitability during
        descent with stochastic samples
      exact_eval: bool, whether to evaluate exploitability during
        descent with exact expectation (req. full payoff tensor)
      avg_trajectory: bool, whether to evaluate w.r.t. the average distribution
        up to time t instead of the distribution at time t
      return_trajectory: bool, whether to record all parameters (e.g., dist)
        during learning and return them -- see solver code for details
    Returns:
      None -- dict of results stored in `results` attribute upon completion
        (key=name of metric, value=[m_0, ..., m_{last_iter}])
    """
    num_players = game.num_players()
    num_strats = game.num_strategies()

    if sym:
      if len(set(num_strats)) != 1:
        raise ValueError('Each player should have the same number of actions.')
      num_strats = num_strats[0]

    params = solver.init_vars(num_strats, num_players)  # dist = params[0]
    if sym:
      dist_avg = np.zeros_like(params[0])
      policies = list(range(num_strats))
      num_ckpts = len(policies)
      form_payoffs_appx = self.construct_payoff_matrices_from_samples_sym
      form_payoffs_exact = self.construct_payoff_matrices_exactly_sym
      exp = sym_exp
      estimate_exploitability = self.estimate_exploitability_sym
    else:
      dist_avg = [np.zeros_like(dist_i) for dist_i in params[0]]
      policies = [list(range(num_strats_i)) for num_strats_i in num_strats]
      num_ckpts = [len(policy_i) for policy_i in policies]
      form_payoffs_appx = self.construct_payoff_matrices_from_samples_nonsym
      form_payoffs_exact = self.construct_payoff_matrices_exactly_nonsym
      exp = nonsym_exp
      estimate_exploitability = self.estimate_exploitability_nonsym

    exps_exact = []
    exps_solver_exact = []
    exps_approx = []
    exps_solver_approx = []
    grad_norms = []

    if return_trajectory:
      params_traj = []

    has_temp = False
    if hasattr(solver, 'temperature') or hasattr(solver, 'p'):
      has_temp = True
      temperatures = []
      if hasattr(solver, 'temperature'):
        temp_attr = 'temperature'
      else:
        temp_attr = 'p'

    early_exit = False

    start = time.time()

    # search for nash (sgd)
    for t in range(num_iterations + 1):
      dist = params[0]
      if return_trajectory:
        params_traj.append(params)

      if return_trajectory:
        params_traj.append(params)

      if has_temp:
        temperatures.append(getattr(solver, temp_attr))

      if num_samples < np.inf:
        payoff_matrices = form_payoffs_appx(game, dist, num_samples,
                                            policies, num_players, num_ckpts)
      else:
        payoff_matrices = form_payoffs_exact(game, dist, num_players)

      grads, exp_sto, exp_solver_sto = solver.compute_gradients(params,
                                                                payoff_matrices)

      if sym:
        grads_dist = grads[0]
        grad_norms.append(simplex.grad_norm(dist, grads_dist))
      else:
        grad_norm = 0.
        grads_dist = grads[0]
        for dist_i, grads_i in zip(dist, grads_dist[0]):
          grad_norm += simplex.grad_norm(dist_i, grads_i)**2.
        grad_norm = np.sqrt(grad_norm)
        grad_norms.append(grad_norm)

      if solver.has_aux:
        solver.record_aux_errors(grads)

      if sym:
        dist_avg += (dist - dist_avg) / float(t + 1)
      else:
        for i, dist_i in enumerate(dist):
          dist_avg[i] += (dist_i - dist_avg[i]) / float(t + 1)

      if avg_trajectory:
        dist_eval = dist_avg
      else:
        dist_eval = dist

      if approx_eval:
        exps_approx.append(exp_sto)
        exps_solver_approx.append(exp_solver_sto)
      if exact_eval:
        pt = game.payoff_tensor()
        exps_exact.append(exp.unreg_exploitability(dist_eval, pt))
        exps_solver_exact.append(solver.exploitability(dist_eval, pt))

      # skip the last update so to avoid computing the matching exploitability
      # and gradient norm information outside the loop
      if t < num_iterations:
        params = solver.update(params, grads, t)
        if misc.isnan(params):
          print('Warning: NaN detected in params post-update. Exiting loop.')
          early_exit = True
          break

    end = time.time()
    solve_runtime = end - start
    start = end

    # evaluating exploitability (monte-carlo)
    exp_estimated = estimate_exploitability(dist_eval, num_eval_samples,
                                            num_ckpts, num_players,
                                            game, policies)

    eval_runtime = time.time() - start

    results = {'exps_approx': exps_approx,
               'exps_solver_approx': exps_solver_approx,
               'exps_exact': exps_exact,
               'exps_solver_exact': exps_solver_exact,
               'exp_estimated': exp_estimated,
               'grad_norms': grad_norms,
               'dist': dist,
               'dist_avg': dist_avg,
               'solve_runtime': solve_runtime,
               'eval_runtime': eval_runtime,
               'early_exit': early_exit}

    if solver.has_aux:
      results.update({'aux_errors': solver.aux_errors})

    if return_trajectory:
      results.update({'params_trajectory': params_traj})

    if has_temp:
      results.update({'temperatures': temperatures})

    self.results = results



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Schulze method.

Based on https://en.wikipedia.org/wiki/Schulze_method.
"""

import functools
import numpy as np
from open_spiel.python.voting import base


class SchulzeVoting(base.AbstractVotingMethod):
  """Implements Schulze's method."""

  def __init__(self):
    pass

  def name(self) -> str:
    return "schulze"

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    alternatives = profile.alternatives
    num_alternatives = profile.num_alternatives()
    pref_mat = profile.pref_matrix()
    strongest_paths = np.zeros(shape=(num_alternatives, num_alternatives),
                               dtype=np.float32)
    # calculating the direct paths
    for i in range(num_alternatives):
      for j in range(num_alternatives):
        if i != j:
          if pref_mat[i, j] > pref_mat[j, i]:
            strongest_paths[i, j] = pref_mat[i, j]
          else:
            strongest_paths[i, j] = 0
    # checking if any indirect paths are better
    for i in range(num_alternatives):
      for j in range(num_alternatives):
        if i != j and strongest_paths[j, i] > 0:
          for k in range(num_alternatives):
            if i != k and j != k:
              # if the path from j to k through i is better, replace
              strongest_paths[j, k] = max(strongest_paths[j, k],
                                          min(strongest_paths[j, i],
                                              strongest_paths[i, k]))

    def compare(x, y):
      return strongest_paths[x, y] - strongest_paths[y, x]
    ranking_idx = np.arange(num_alternatives)
    sorted_ranking_idx = sorted(ranking_idx, key=functools.cmp_to_key(compare),
                                reverse=True)
    # Define the scores as the sum of preferences for everything it beats in
    # the order.
    cumul_score = 0
    # start at the end and work backwards
    ranking_alts = [alternatives[sorted_ranking_idx[-1]]]
    scores = [0]
    i = num_alternatives - 2
    while i >= 0:
      alt_idx_i = sorted_ranking_idx[i]
      alt_idx_j = sorted_ranking_idx[i+1]
      ranking_alts.insert(0, alternatives[alt_idx_i])
      cumul_score += pref_mat[alt_idx_i, alt_idx_j]
      scores.insert(0, cumul_score)
      i -= 1
    return base.RankOutcome(rankings=ranking_alts, scores=scores)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluation."""

from absl import flags
import jax
import jax.numpy as jnp
import numpy as np

FLAGS = flags.FLAGS


@jax.jit
def compute_best_response_strategy(utility):
  actions_count = utility.shape[-1]
  opponent_action = jnp.argmin(utility, axis=-1)
  opponent_strategy = jax.nn.one_hot(opponent_action, actions_count)
  return opponent_strategy


@jax.jit
def compute_values_against_best_response(strategy, payoff):
  utility = jnp.matmul(strategy, payoff)
  br_strategy = compute_best_response_strategy(utility)
  return jnp.matmul(payoff, jnp.transpose(br_strategy))


def evaluate_against_best_response(agent, payoff_batch, steps_count):
  """Evaluation against best response agent.

  Args:
    agent: Agent model.
    payoff_batch: Payoff matrix.
    steps_count: Number of steps.
  """
  current_policy = agent.initial_policy()
  values = jax.vmap(compute_values_against_best_response)(current_policy,
                                                          payoff_batch)
  for step in range(steps_count):
    current_policy = agent.next_policy(values)
    values = jax.vmap(compute_values_against_best_response)(current_policy,
                                                            payoff_batch)
    values = jnp.transpose(values, [0, 1, 2])
    value = jnp.matmul(current_policy, values)

    for i in range(value.shape[0]):
      print(step, np.mean(np.asarray(value[i])))


def compute_regrets(payoff_batch, strategy_x, strategy_y):
  values_y = -jnp.matmul(strategy_x, payoff_batch)
  values_x = jnp.transpose(
      jnp.matmul(payoff_batch, jnp.transpose(strategy_y, [0, 2, 1])), [0, 2, 1])
  value_x = jnp.matmul(
      jnp.matmul(strategy_x, payoff_batch),
      jnp.transpose(strategy_y, [0, 2, 1]))
  value_y = -value_x
  regrets_x = values_x - value_x
  regrets_y = values_y - value_y
  return regrets_x, regrets_y


def evaluate_in_selfplay(agent_x, agent_y, payoff_batch, steps_count):
  """Evalute in selfplay.

  Args:
    agent_x: First agent.
    agent_y: Second agent.
    payoff_batch: Payoff matrix.
    steps_count: Number of steps.
  """
  payoff_batch_size = payoff_batch.shape[0]

  regret_sum_x = np.zeros(shape=[payoff_batch_size, 1, FLAGS.num_actions])
  regret_sum_y = np.zeros(shape=[payoff_batch_size, 1, FLAGS.num_actions])
  strategy_x = agent_x.initial_policy()
  strategy_y = agent_y.initial_policy()

  regrets_x, regrets_y = compute_regrets(payoff_batch, strategy_x, strategy_y)
  regret_sum_x += regrets_x
  regret_sum_y += regrets_y
  for s in range(steps_count):
    values_y = -jnp.matmul(strategy_x, payoff_batch)
    values_x = jnp.transpose(
        jnp.matmul(payoff_batch, jnp.transpose(strategy_y, [0, 2, 1])),
        [0, 2, 1])

    values_x = jnp.transpose(values_x, [0, 2, 1])
    values_y = jnp.transpose(values_y, [0, 2, 1])
    strategy_x = agent_x.next_policy(values_x)
    strategy_y = agent_y.next_policy(values_y)

    regrets_x, regrets_y = compute_regrets(payoff_batch, strategy_x, strategy_y)
    regret_sum_x += regrets_x
    regret_sum_y += regrets_y
    print(
        jnp.mean(
            jnp.max(
                jnp.concatenate([regret_sum_x, regret_sum_y], axis=2),
                axis=[1, 2]) / (s + 1)))



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantal Response Equilibrium (QRE) Stochastic Approximate Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np
from scipy import special

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability as exp


class Solver(object):
  """QRE Solver without auxiliary y variable."""

  def __init__(self, temperature=1., proj_grad=True, euclidean=False,
               cheap=False, lrs=(1e-2,), exp_thresh=-1., vr=True,
               rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    if temperature < 0.:
      raise ValueError("temperature must be non-negative")
    self.num_players = None
    self.temperature = temperature
    self.proj_grad = proj_grad
    self.cheap = cheap
    self.vr = vr
    self.pm_vr = None
    self.rnd_init = rnd_init
    self.lrs = lrs
    self.exp_thresh = exp_thresh
    self.has_aux = False

    self.euclidean = euclidean
    if euclidean:
      self.update = self.euc_descent_step
    else:
      self.update = self.mirror_descent_step

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if self.rnd_init:
      init_dist = self.random.rand(num_strats)
    else:
      init_dist = np.ones(num_strats)
    init_dist /= init_dist.sum()
    init_anneal_steps = 0
    if self.cheap and self.vr:
      self.pm_vr = np.zeros((num_strats, num_strats))
    return (init_dist, init_anneal_steps)

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients (and exploitabilities) for all parameters.

    Args:
      params: tuple of params (dist, anneal_steps), see gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      tuple of gradients (grad_dist, grad_anneal_steps), see gradients
      unregularized exploitability (stochastic estimate)
      tsallis regularized exploitability (stochastic estimate)
    """
    if self.cheap and self.vr:
      grads, pm_vr, exp_sto, exp_solver_sto = self.cheap_gradients_vr(
          self.random, *params, payoff_matrices, self.num_players, self.pm_vr,
          self.temperature, self.proj_grad,)
      self.pm_vr = pm_vr
      return grads, exp_sto, exp_solver_sto
    elif self.cheap and not self.vr:
      return self.cheap_gradients(self.random, *params, payoff_matrices,
                                  self.num_players, self.temperature,
                                  self.proj_grad)
    else:
      return self.gradients(*params, payoff_matrices, self.num_players,
                            self.temperature, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Compute and return shannon entropy regularized exploitability.

    Args:
      params: tuple of params (dist, y), see qre.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability of current dist
    """
    return exp.qre_exploitability(params, payoff_matrices, self.temperature)

  def gradients(self, dist, anneal_steps, payoff_matrices, num_players,
                temperature=0., proj_grad=True):
    """Computes exploitablity gradient and aux variable gradients.

    Args:
      dist: 1-d np.array, current estimate of nash distribution
      anneal_steps: int, elapsed num steps since last anneal
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
      num_players: int, number of players, in case payoff_matrices is
        abbreviated
      temperature: non-negative float, default 0.
      proj_grad: bool, if True, projects dist gradient onto simplex
    Returns:
      gradient of exploitability w.r.t. (dist, anneal_steps) as tuple
      unregularized exploitability (stochastic estimate)
      tsallis regularized exploitability (stochastic estimate)
    """
    y = nabla = payoff_matrices[0].dot(dist)
    if temperature >= 1e-3:
      br = special.softmax(y / temperature)
      br_mat = (np.diag(br) - np.outer(br, br)) / temperature
      log_br_safe = np.clip(np.log(br), -1e5, 0)
      br_policy_gradient = nabla - temperature * (log_br_safe + 1)
    else:
      power = np.inf
      s = np.linalg.norm(y, ord=power)
      br = np.zeros_like(dist)
      maxima = (y == s)
      br[maxima] = 1. / maxima.sum()
      br_mat = np.zeros((br.size, br.size))
      br_policy_gradient = np.zeros_like(br)

    unreg_exp = np.max(y) - y.dot(dist)
    entr_br = temperature * special.entr(br).sum()
    entr_dist = temperature * special.entr(dist).sum()
    reg_exp = y.dot(br - dist) + entr_br - entr_dist

    policy_gradient = np.array(nabla)
    if temperature > 0:
      log_dist_safe = np.clip(np.log(dist), -1e5, 0)
      policy_gradient -= temperature * (log_dist_safe + 1)
    other_player_fx = (br - dist) + br_mat.dot(br_policy_gradient)

    other_player_fx_translated = payoff_matrices[1].dot(other_player_fx)
    grad_dist = -policy_gradient
    grad_dist += (num_players - 1) * other_player_fx_translated
    if proj_grad:
      grad_dist = simplex.project_grad(grad_dist)

    if reg_exp < self.exp_thresh:
      self.temperature = np.clip(temperature / 2., 0., np.inf)
      if self.temperature < 1e-3:
        self.temperature = 0.
      grad_anneal_steps = -anneal_steps
    else:
      grad_anneal_steps = 1

    return (grad_dist, grad_anneal_steps), unreg_exp, reg_exp

  def cheap_gradients(self, random, dist, anneal_steps, payoff_matrices,
                      num_players, temperature=0., proj_grad=True):
    """Computes exploitablity gradient and aux variable gradients with samples.

    This implementation takes payoff_matrices as input so technically uses
    O(d^2) compute but only a single column of payoff_matrices is used to
    perform the update so can be re-implemented in O(d) if needed.

    Args:
      random: random number generator, np.random.RandomState(seed)
      dist: 1-d np.array, current estimate of nash distribution
      anneal_steps: int, elapsed num steps since last anneal
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
      num_players: int, number of players, in case payoff_matrices is
        abbreviated
      temperature: non-negative float, default 0.
      proj_grad: bool, if True, projects dist gradient onto simplex
    Returns:
      gradient of exploitability w.r.t. (dist, anneal_steps) as tuple
      unregularized exploitability (stochastic estimate)
      tsallis regularized exploitability (stochastic estimate)
    """
    del anneal_steps
    action_1 = random.choice(dist.size, p=dist)
    y = nabla = payoff_matrices[0][:, action_1]
    if temperature >= 1e-3:
      br = special.softmax(y / temperature)
      br_mat = (np.diag(br) - np.outer(br, br)) / temperature
      br_policy_gradient = nabla - temperature * (np.log(br) + 1)
    else:
      power = np.inf
      s = np.linalg.norm(y, ord=power)
      br = np.zeros_like(dist)
      maxima = (y == s)
      br[maxima] = 1. / maxima.sum()
      br_mat = np.zeros((br.size, br.size))
      br_policy_gradient = np.zeros_like(br)

    unreg_exp = np.max(y) - y.dot(dist)
    entr_br = temperature * special.entr(br).sum()
    entr_dist = temperature * special.entr(dist).sum()
    reg_exp = y.dot(br - dist) + entr_br - entr_dist

    policy_gradient = nabla - temperature * (np.log(dist) + 1)
    other_player_fx = (br - dist) + br_mat.dot(br_policy_gradient)

    action_u = random.choice(dist.size)  # uniform, ~importance sampling
    other_player_fx = dist.size * other_player_fx[action_u]
    other_player_fx_translat = payoff_matrices[1, :, action_u] * other_player_fx
    grad_dist = -policy_gradient + (num_players - 1) * other_player_fx_translat
    if proj_grad:
      grad_dist = simplex.project_grad(grad_dist)

    return (grad_dist, None), unreg_exp, reg_exp

  def cheap_gradients_vr(self, random, dist, anneal_steps, payoff_matrices,
                         num_players, pm_vr, temperature=0., proj_grad=True,
                         version=0):
    """Computes exploitablity gradient and aux variable gradients with samples.

    This implementation takes payoff_matrices as input so technically uses
    O(d^2) compute but only a single column of payoff_matrices is used to
    perform the update so can be re-implemented in O(d) if needed.

    Args:
      random: random number generator, np.random.RandomState(seed)
      dist: 1-d np.array, current estimate of nash distribution
      anneal_steps: int, elapsed num steps since last anneal
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
      num_players: int, number of players, in case payoff_matrices is
        abbreviated
      pm_vr: approximate payoff_matrix for variance reduction
      temperature: non-negative float, default 0.
      proj_grad: bool, if True, projects dist gradient onto simplex
      version: int, default 0, two options for variance reduction
    Returns:
      gradient of exploitability w.r.t. (dist, anneal_steps) as tuple
      unregularized exploitability (stochastic estimate)
      tsallis regularized exploitability (stochastic estimate)
    """
    del anneal_steps
    if pm_vr is None:
      raise ValueError("pm_vr must be np.array of shape (num_strats,) * 2")
    if (not isinstance(version, int)) or (version < 0) or (version > 1):
      raise ValueError("version must be non-negative int < 2")

    action_1 = random.choice(dist.size, p=dist)
    y = nabla = payoff_matrices[0][:, action_1]
    if temperature >= 1e-3:
      br = special.softmax(y / temperature)
      br_mat = (np.diag(br) - np.outer(br, br)) / temperature
      br_policy_gradient = nabla - temperature * (np.log(br) + 1)
    else:
      power = np.inf
      s = np.linalg.norm(y, ord=power)
      br = np.zeros_like(dist)
      maxima = (y == s)
      br[maxima] = 1. / maxima.sum()
      br_mat = np.zeros((br.size, br.size))
      br_policy_gradient = np.zeros_like(br)

    unreg_exp = np.max(y) - y.dot(dist)
    entr_br = temperature * special.entr(br).sum()
    entr_dist = temperature * special.entr(dist).sum()
    reg_exp = y.dot(br - dist) + entr_br - entr_dist

    policy_gradient = nabla - temperature * (np.log(dist) + 1)
    other_player_fx = (br - dist) + br_mat.dot(br_policy_gradient)

    if version == 0:
      other_player_fx_translated = pm_vr.dot(other_player_fx)
      action_u = random.choice(dist.size)  # uniform, ~importance sampling
      other_player_fx = other_player_fx[action_u]
      m = dist.size
      pm_mod = m * (payoff_matrices[1, :, action_u] - pm_vr[:, action_u])
      other_player_fx_translated += pm_mod * other_player_fx
    elif version == 1:
      other_player_fx_translated = np.sum(pm_vr, axis=1)
      action_u = random.choice(dist.size)  # uniform, ~importance sampling
      other_player_fx = other_player_fx[action_u]
      pm_mod = dist.size * payoff_matrices[1, :, action_u]
      r = dist.size * pm_vr[:, action_u]
      other_player_fx_translated += pm_mod * other_player_fx - r

    grad_dist = -policy_gradient
    grad_dist += (num_players - 1) * other_player_fx_translated
    if proj_grad:
      grad_dist = simplex.project_grad(grad_dist)

    if version == 0:
      pm_vr[:, action_u] = payoff_matrices[1, :, action_u]
    elif version == 1:
      pm_vr[:, action_u] = payoff_matrices[1, :, action_u] * other_player_fx

    return (grad_dist, None), pm_vr, unreg_exp, reg_exp

  def euc_descent_step(self, params, grads, t):
    """Projected gradient descent on exploitability using Euclidean projection.

    Args:
      params: tuple of variables to be updated (dist, anneal_steps)
      grads: tuple of variable gradients (grad_dist, grad_anneal_steps)
      t: int, solver iteration
    Returns:
      new_params: tuple of update params (new_dist, new_anneal_steps)
    """
    del t
    lr_dist = self.lrs[0]
    new_params = [params[0] - lr_dist * grads[0]]
    new_params = euc_project(*new_params)
    new_params += (params[1] + grads[1],)
    return new_params

  def mirror_descent_step(self, params, grads, t):
    """Entropic mirror descent on exploitability.

    Args:
      params: tuple of variables to be updated (dist, anneal_steps)
      grads: tuple of variable gradients (grad_dist, grad_anneal_steps)
      t: int, solver iteration
    Returns:
      new_params: tuple of update params (new_dist, new_anneal_steps)
    """
    del t
    lr_dist = self.lrs[0]
    new_params = [np.log(np.clip(params[0], 0, np.inf)) - lr_dist * grads[0]]
    new_params = mirror_project(*new_params)
    new_params += (params[1] + grads[1],)
    return new_params


def euc_project(dist):
  """Project variables onto their feasible sets (euclidean proj for dist).

  Args:
    dist: 1-d np.array, current estimate of nash distribution
  Returns:
    projected variables (dist,) as tuple
  """
  dist = simplex.euclidean_projection_onto_simplex(dist)

  return (dist,)


def mirror_project(dist):
  """Project variables onto their feasible sets (softmax for dist).

  Args:
    dist: 1-d np.array, current estimate of nash distribution
  Returns:
    projected variables (dist,) as tuple
  """
  dist = special.softmax(dist)

  return (dist,)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A pyspiel config for meta-generated fruit trading games.
"""

import collections

from ml_collections import config_dict

from open_spiel.python.games.chat_games.envs.base_envs import trade_fruit_with_info as env_trade_fruit_with_info
from open_spiel.python.games.chat_games.envs.observations import summary
from open_spiel.python.games.chat_games.envs.observations import utils as obs_utils
from open_spiel.python.games.chat_games.envs.payoffs import trade_fruit as payoffs_trade_fruit
from open_spiel.python.games.chat_games.envs.scenarios.domains import trade_fruit as scenario_trade_fruit
from open_spiel.python.games.chat_games.envs.scenarios.players import names as names_trade_fruit


def get_config():
  """Get configuration for chat game."""
  config = config_dict.ConfigDict()

  num_players = 2

  observations = [
      obs_utils.Observation(summary.PREFIX, summary.POSTFIX)
      for _ in range(num_players)
  ]

  header = env_trade_fruit_with_info.HEADER

  payoffs = [payoffs_trade_fruit.PAYOFF]

  examples_names = names_trade_fruit.NAMES

  examples_private_info = collections.OrderedDict()
  examples_private_info['fruit_endowment'] = [scenario_trade_fruit.ENDOWMENT_A,
                                              scenario_trade_fruit.ENDOWMENT_B]
  examples_private_info['fruit_valuations'] = [scenario_trade_fruit.VALUATION_A,
                                               scenario_trade_fruit.VALUATION_B]

  scenario_a = env_trade_fruit_with_info.Scenario(
      scenario_trade_fruit.SCENARIO_A,
      'Bob',
      'Suzy',
      scenario_trade_fruit.ENDOWMENT_A,
      scenario_trade_fruit.VALUATION_A)
  scenario_b = env_trade_fruit_with_info.Scenario(
      scenario_trade_fruit.SCENARIO_B,
      'Jill',
      'George',
      scenario_trade_fruit.ENDOWMENT_B,
      scenario_trade_fruit.VALUATION_B)
  examples_scenarios = [scenario_a, scenario_b]

  llm_termination_prompt = scenario_trade_fruit.LLM_TERMINATION_PROMPT

  params = {'num_distinct_actions': num_players,
            'num_llm_seeds': 2,
            'num_init_states': 3,
            'num_players': num_players,
            'min_utility': min([float(p.min) for p in payoffs]),
            'max_utility': max([float(p.max) for p in payoffs]),
            'num_max_replies': 3}

  config.params = params

  config.game = config_dict.ConfigDict()
  config.game.observations = observations
  config.game.header = header
  config.game.payoffs = payoffs
  config.game.num_names = 10
  config.game.num_private_info = (3, 3)
  config.game.examples_names = examples_names
  config.game.examples_private_info = examples_private_info
  config.game.examples_scenarios = examples_scenarios
  config.game.llm_list_suffix = 'Output: '
  config.game.llm_termination_prompt = llm_termination_prompt

  return config



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for bargaining."""


from absl.testing import absltest

import pyspiel
barg = pyspiel.bargaining


class GamesBargainingTest(absltest.TestCase):

  def test_constants(self):
    self.assertEqual(barg.NumItemTypes, 3)
    self.assertEqual(barg.PoolMinNumItems, 5)
    self.assertEqual(barg.PoolMaxNumItems, 7)
    self.assertEqual(barg.TotalValueAllItems, 10)

  def test_game_specific_constants(self):
    game0 = pyspiel.load_game("bargaining")
    self.assertEqual(game0.max_turns(), 10)
    self.assertEqual(game0.discount(), 1.0)
    self.assertEqual(game0.prob_end(), 0.0)

    game1 = pyspiel.load_game(
        "bargaining(max_turns=15,discount=0.9,prob_end=0.1)"
    )
    self.assertEqual(game1.max_turns(), 15)
    self.assertEqual(game1.discount(), 0.9)
    self.assertEqual(game1.prob_end(), 0.1)

  def test_game_mechanism(self):
    game = pyspiel.load_game("bargaining")
    state = game.new_initial_state()

    # first check the instance matches the true instance
    true_instance = [(1, 2, 3), (8, 1, 0), (4, 0, 2)]
    state.apply_action(0)
    cur_instance = state.instance()
    cur_instance = [
        tuple(cur_instance.pool),
        tuple(cur_instance.values[0]),
        tuple(cur_instance.values[1])
    ]

    for item1, item2 in zip(true_instance, cur_instance):
      self.assertEqual(item1, item2)

    # then set a new instance and check it works
    all_instances = game.all_instances()
    new_instance = all_instances[2]
    state.set_instance(new_instance)
    new_instance = [
        tuple(new_instance.pool),
        tuple(new_instance.values[0]),
        tuple(new_instance.values[1])
    ]
    cur_instance = state.instance()
    cur_instance = [
        tuple(cur_instance.pool),
        tuple(cur_instance.values[0]),
        tuple(cur_instance.values[1])
    ]
    for item1, item2 in zip(cur_instance, new_instance):
      self.assertEqual(item1, item2)

  def test_offer_and_instance_map(self):
    game = pyspiel.load_game("bargaining")
    all_offers = game.all_offers()
    for i, offer in enumerate(all_offers):
      self.assertEqual(game.get_offer_index(offer), i)
    for i, instance in enumerate(game.all_instances()):
      self.assertEqual(game.get_instance_index(instance), i)

  def test_get_possible_opponent_values(self):
    game = pyspiel.load_game("bargaining")
    self.assertEqual(
        game.get_possible_opponent_values(0, [1, 2, 3], [8, 1, 0]),
        [
            [4, 0, 2],
            [7, 0, 1],
            [1, 3, 1],
        ],
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for nash conv."""

from absl.testing import absltest

from open_spiel.python import policy
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.games import crowd_modelling
import pyspiel


class BestResponseTest(absltest.TestCase):

  def test_python_game(self):
    """Checks if the NashConv is consistent through time."""
    game = crowd_modelling.MFGCrowdModellingGame()
    uniform_policy = policy.UniformRandomPolicy(game)
    nash_conv_fp = nash_conv.NashConv(game, uniform_policy)

    self.assertAlmostEqual(nash_conv_fp.nash_conv(), 2.8135365543870385)

  def test_cpp_game(self):
    """Checks if the NashConv is consistent through time."""
    game = pyspiel.load_game("mfg_crowd_modelling")
    uniform_policy = policy.UniformRandomPolicy(game)
    nash_conv_fp = nash_conv.NashConv(game, uniform_policy)

    self.assertAlmostEqual(nash_conv_fp.nash_conv(), 2.8135365543870385)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A communication format (substrate) for setting schedules.
"""

from open_spiel.python.games.chat_games.envs.utils import text


CHAR_OPT = '%'
CHAR_MSG = '#'
BLOCK_LEN = 28

SPECIAL_CHARS = (CHAR_OPT, CHAR_MSG)
BLOCK_OPT = CHAR_OPT * BLOCK_LEN
BLOCK_MSG = CHAR_MSG * BLOCK_LEN

PLAIN = ('\n\n' + BLOCK_MSG + '\n' +
         'Schedule Proposal Message:\n' +
         'from: {sender}\n' +
         'to: {receiver}\n' +
         BLOCK_MSG + '\n\n')

W_OPTS_PREFIX = '\n\n' + BLOCK_OPT + '\n\n'


def strip_msg(msg: str, terminal_str: str = '') -> str:
  return text.strip_msg(msg, BLOCK_MSG, BLOCK_OPT, terminal_str)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Policy Gradient (PG)."""

from absl import logging  # pylint:disable=unused-import

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import updates


class Solver(updates.Solver):
  """PG Solver."""

  def __init__(self, proj_grad=True, euclidean=False, lrs=(1e-1,),
               rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    super().__init__(proj_grad, euclidean, rnd_init, seed)
    self.lrs = lrs

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients for all parameters.

    Args:
      params: tuple of params (dist,), see pg.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      tuple of gradients (grad_dist,), see pg.gradients
      unregularized exploitability (stochastic estimate)
      unregularized exploitability (stochastic estimate) *duplicate
    """
    return gradients(*params, payoff_matrices, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Policy gradient does not minimize any exploitability so return NaN.

    Args:
      params: tuple of params (dist,)
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      np.NaN
    """
    return np.nan


def gradients(dist, payoff_matrices, proj_grad=True):
  """Computes exploitablity gradient.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of payoff w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    unregularized exploitability (stochastic estimate) *duplicate
  """
  nabla = payoff_matrices[0].dot(dist)

  unreg_exp = np.max(nabla) - nabla.dot(dist)

  grad_dist = -nabla
  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)

  return (grad_dist,), unreg_exp, unreg_exp



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Elo rating system.

There are two ways to compute the Elo ratings:
  1. From a win matrix and a draw matrix.
  2. From a list of match records.

The function here is a wrapper to method (1) that simply converts numpy arrays
to lists of lists and then calls the pybind11 wrapper. For examples of (2),
please see elo_test.py.

Both are wrappers around the C++ implementation in evaluation/elo.cc, which is
based on the algorithm of Hunter, MM Algorithms for Generalized Bradley-Terry
Models, The Annals of Statistics 2004, Vol. 32, No. 1, 384--406.
"""

from typing import Optional
import numpy as np
import pyspiel


def compute_ratings_from_matrices(
    win_matrix: np.ndarray,
    draw_matrix: Optional[np.ndarray] = None,
    smoothing_factor: float = pyspiel.elo.DEFAULT_SMOOTHING_FACTOR,
    max_iterations: int = pyspiel.elo.DEFAULT_MAX_ITERATIONS,
    convergence_delta: float = pyspiel.elo.DEFAULT_CONVERGENCE_DELTA,
) -> np.ndarray:
  """Compute Elo ratings from a win matrix and a draw matrix."""
  options = pyspiel.elo.default_elo_options()
  options.smoothing_factor = smoothing_factor
  options.max_iterations = max_iterations
  options.convergence_delta = convergence_delta
  return pyspiel.elo.compute_ratings_from_matrices(
      win_matrix=win_matrix.tolist(),
      draw_matrix=(draw_matrix.tolist() if draw_matrix is not None else []),
      options=options,
  )



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Collection of useful functions and classes."""

from typing import List, Optional

import numpy as np

from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import distribution as distribution_std
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import policy_value
import pyspiel


class MergedPolicy(policy_std.Policy):
  """Merge several policies."""

  def __init__(
      self,
      game,
      player_ids,
      policies: List[policy_std.Policy],
      weights: List[float],
      distributions: Optional[List[distribution_std.Distribution]] = None,
  ):
    """Initializes the merged policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
      policies: A `List[policy_std.Policy]` object.
      weights: A `List[float]` object. They should sum to 1.
      distributions: A `List[distribution_std.Distribution]` object.
    """
    super(MergedPolicy, self).__init__(game, player_ids)
    self._policies = policies
    self._distributions = distributions
    self._weights = weights
    if distributions is None:
      distributions = [
          distribution.DistributionPolicy(game, policy) for policy in policies
      ]
    else:
      assert len(policies) == len(
          distributions
      ), f'Length mismatch {len(policies)} != {len(distributions)}'
    assert len(policies) == len(
        weights
    ), f'Length mismatch {len(policies)} != {len(weights)}'

  def action_probabilities(self, state, player_id=None):
    action_prob = []
    legal = state.legal_actions()
    num_legal = len(legal)
    for a in legal:
      merged_pi = 0.0
      norm_merged_pi = 0.0
      for p, d, w in zip(self._policies, self._distributions, self._weights):
        merged_pi += w * d(state) * p(state)[a]
        norm_merged_pi += w * d(state)
      if norm_merged_pi > 0.0:
        action_prob.append((a, merged_pi / norm_merged_pi))
      else:
        action_prob.append((a, 1.0 / num_legal))
    return dict(action_prob)


class MixedDistribution:
  """Mixes a list of distributions wrt. a list of weights.

  The mixed distribution remains a probability distribution over states.

  Attributes:
    mus: The state distributions being mixed.
    weights: The list of weights of each `mus` member.
    _mus: The state distributions being mixed, post-pruning.
    _weights: The list of weights of each `mus` member, post-pruning.
    _tol: Tolerance (`mus` members with weights below tolerance are ignored)
    _value_str_cache: Cache for value_str calls.
  """

  def __init__(self, mus, weights, tol=1e-4):
    """Mixes the distribution.

    Args:
      mus: List of distributions to mix.
      weights: List of weights to mix `mus` over.
      tol: Tolerance (`mus` members with weights below tolerance are ignored)
    """
    self.mus = mus
    self.weights = weights
    self._tol = tol
    self._prune()
    self._value_str_cache = {}

  def _prune(self):
    self._mus = [mu for w, mu in zip(self.weights, self.mus) if w > self._tol]
    self._weights = [w for w in self.weights if w > self._tol]
    self._weights = [w / sum(self._weights) for w in self._weights]

  def value(self, state):
    """Returns the probability of the distribution on the state.

    Args:
      state: A `pyspiel.State` object.

    Returns:
      A `float`.
    """
    return sum([
        weight * mu.value(state) for weight, mu in zip(self._weights, self._mus)
    ])

  def value_str(self, state_str, default_value=None):
    """Returns the probability of the distribution on the given state string.

    Args:
      state_str: A string.
      default_value: If not None, return this value if the state is not in the
        support of the distribution.

    Returns:
      A `float`.
    """
    if state_str not in self._value_str_cache:
      self._value_str_cache[state_str] = sum([
          weight * mu.value_str(state_str, default_value)
          for weight, mu in zip(self._weights, self._mus)
      ])
    return self._value_str_cache[state_str]

  def __call__(self, state):
    """Turns the distribution into a callable.

    Args:
      state: The current state of the game.

    Returns:
      Float: probability.
    """
    return self.value(state)


def get_exact_value(
    pi: policy_std.Policy, mu: distribution_std.Distribution, game
):
  """Computes the exact value of playing `pi` against distribution `mu`.

  Args:
    pi: A policy object whose value is evaluated against `mu`.
    mu: A distribution object against which `pi` is evaluated.
    game: A pyspiel.Game object, the evaluation game.

  Returns:
    Exact value of `pi` in `game` against `mu`.
  """
  root_state = game.new_initial_states()[0]
  return policy_value.PolicyValue(game, mu, pi).value(root_state)


def sample_value(
    pi: policy_std.Policy, mu: distribution_std.Distribution, game
):
  """Samples the value of playing `pi` against distribution `mu`.

  Args:
    pi: A policy object whose value is evaluated against `mu`.
    mu: A distribution object against which `pi` is evaluated.
    game: A pyspiel.Game object, the evaluation game.

  Returns:
    Sampled value of `pi` in `game` against `mu`.
  """
  mfg_state = game.new_initial_states()[0]
  total_reward = 0.0
  while not mfg_state.is_terminal():
    if mfg_state.current_player() == pyspiel.PlayerId.CHANCE:
      action_list, prob_list = zip(*mfg_state.chance_outcomes())
      action = np.random.choice(action_list, p=prob_list)
      mfg_state.apply_action(action)
    elif mfg_state.current_player() == pyspiel.PlayerId.MEAN_FIELD:
      dist_to_register = mfg_state.distribution_support()
      dist = [mu.value_str(str_state, 0.0) for str_state in dist_to_register]
      mfg_state.update_distribution(dist)
    else:
      total_reward += mfg_state.rewards()[0]
      action_prob = pi(mfg_state)
      action = np.random.choice(
          list(action_prob.keys()), p=list(action_prob.values())
      )
      mfg_state.apply_action(action)

  return total_reward


def get_nu_values(policies, nu, game):
  rewards = np.zeros(len(policies))
  mu = distribution.DistributionPolicy(
      game, MergedPolicy(game, None, policies, nu)
  )
  for index, policy in enumerate(policies):
    rewards[index] = sample_value(policy, mu, game)
  return rewards



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Examples of fruit trading scenarios -- useful for generating more examples.
"""

from open_spiel.python.games.chat_games.envs.observations import summary
from open_spiel.python.games.chat_games.envs.termination import utils as term_utils
from open_spiel.python.games.chat_games.envs.utils import text

# Scenario A
SCENARIO_A_LIST = ['Hi {receiver},',
                   'I would like to trade you 1 banana for 1 apple.',
                   'Would you like to trade with me?',
                   'Best,', '{sender}']
SCENARIO_A = '\n\n'.join(text.wrap(SCENARIO_A_LIST))

ENDOWMENT_A_LIST = ['apple: 1', 'banana: 2', 'blueberry: 0', 'kiwi: 0']
ENDOWMENT_A = '\n'.join(text.wrap(ENDOWMENT_A_LIST))

VALUATION_A_LIST = ['apple: 10',
                    'banana: 5',
                    'blueberry: 1',
                    'kiwi: 3']
VALUATION_A = '\n'.join(text.wrap(VALUATION_A_LIST))

# Scenario B
SCENARIO_B_LIST = ['Hi {receiver},',
                   'I would like to trade you 3 blueberries for 1 banana.',
                   'Would you like to trade with me?',
                   'Best,', '{sender}']
SCENARIO_B = '\n\n'.join(text.wrap(SCENARIO_A_LIST))

ENDOWMENT_B_LIST = ['apple: 0', 'banana: 0', 'blueberry: 5', 'kiwi: 3']
ENDOWMENT_B = '\n'.join(text.wrap(ENDOWMENT_B_LIST))

VALUATION_B_LIST = ['apple: 8',
                    'banana: 7',
                    'blueberry: 2',
                    'kiwi: 2']
VALUATION_B = '\n'.join(text.wrap(VALUATION_B_LIST))

query = ('Read the following summary of a dialgoue between two parties ' +
         'attempting to reach a trade agreement. Have the players reached a ' +
         'trade agreement? If a trade has been accepted or the players cannot' +
         ' come to an agreement, respond Yes. Otherwise, if the players are ' +
         'still discussing terms, respond No.' +
         'Here is the dialogue:\n\n{msg}\n\n' + '&' *50 +
         'Response: ')

LLM_TERMINATION_PROMPT = term_utils.Termination(query,
                                                summary.PREFIX,
                                                summary.POSTFIX)
